from __future__ import annotations

import json
from pathlib import Path
from types import SimpleNamespace

import pytest
from typer.testing import CliRunner

from worai.commands import graph as graph_cmd
from worai.errors import UsageError


def _make_url_result(url: str, errors: list | None = None, warnings: list | None = None) -> SimpleNamespace:
    errors = errors or []
    warnings = warnings or []
    return SimpleNamespace(
        url=url,
        errors=[SimpleNamespace(message=e, focus_node="urn:focus") for e in errors],
        warnings=[SimpleNamespace(message=w, focus_node="urn:focus") for w in warnings],
        error_count=len(errors),
        warning_count=len(warnings),
        google_merchant=SimpleNamespace(eligible=len(errors) == 0),
    )


class _FakeReport:
    def __init__(
        self,
        load_errors: list | None = None,
        text: str = "OK",
        data: dict | None = None,
        url_results: list | None = None,
    ) -> None:
        self.load_errors = load_errors or []
        self._text = text
        self._data = data or {"summary": "ok"}
        self.schema_compliance = SimpleNamespace(by_url=url_results or [])

    def to_text(self) -> str:
        return self._text

    def to_dict(self) -> dict:
        return self._data


class _FakeAuditor:
    def __init__(self, report: _FakeReport) -> None:
        self._report = report

    def audit(self, path: str, options: object) -> _FakeReport:
        return self._report


def _patch_sdk(monkeypatch, report: _FakeReport | None = None, auditor_cls=None) -> None:
    if report is None:
        report = _FakeReport()
    if auditor_cls is None:
        _auditor = _FakeAuditor(report)

        class _Cls:
            def __call__(self):
                return _auditor

        auditor_cls = _Cls()

    monkeypatch.setattr(graph_cmd, "_GraphAuditor", auditor_cls)
    monkeypatch.setattr(graph_cmd, "_AuditOptions", SimpleNamespace)


# ---------------------------------------------------------------------------
# _render_audit_text — failures-only schema compliance
# ---------------------------------------------------------------------------

_COMPLIANCE_SECTION = "--- Schema Compliance ---"


def _report_with_urls(ok_urls: list[str], fail_urls: list[str]) -> _FakeReport:
    url_results = [_make_url_result(u) for u in ok_urls] + [
        _make_url_result(u, errors=["Missing required field"]) for u in fail_urls
    ]
    body = f"=== Graph Audit Report ===\n\n{_COMPLIANCE_SECTION}\n"
    return _FakeReport(text=body, url_results=url_results)


def _compliance_report(ok_urls: list[str], fail_urls: list[str]) -> _FakeReport:
    url_results = [_make_url_result(u) for u in ok_urls] + [
        _make_url_result(u, errors=["Missing required field"]) for u in fail_urls
    ]
    body = f"=== Graph Audit Report ===\n\n{_COMPLIANCE_SECTION}\n"
    return _FakeReport(text=body, url_results=url_results)


def test_render_audit_text_default_shows_summary_only() -> None:
    report = _compliance_report(ok_urls=["https://example.com/ok"], fail_urls=["https://example.com/fail"])
    rendered = graph_cmd._render_audit_text(report)
    assert "URLs checked : 2" in rendered
    assert "URLs failing : 1" in rendered
    assert "https://example.com/ok" not in rendered
    assert "https://example.com/fail" not in rendered
    assert "[FAIL]" not in rendered


def test_render_audit_text_show_url_violations_lists_failing_urls() -> None:
    report = _compliance_report(ok_urls=["https://example.com/ok"], fail_urls=["https://example.com/fail"])
    rendered = graph_cmd._render_audit_text(report, show_url_violations=True)
    assert "https://example.com/ok" not in rendered
    assert "https://example.com/fail" in rendered
    assert "[FAIL]" in rendered
    assert "[OK]" not in rendered


def test_render_audit_text_no_failures_shows_zero_count() -> None:
    report = _compliance_report(ok_urls=["https://example.com/a", "https://example.com/b"], fail_urls=[])
    rendered = graph_cmd._render_audit_text(report)
    assert "URLs checked : 2" in rendered
    assert "URLs failing : 0" in rendered
    assert "[FAIL]" not in rendered


def test_render_audit_text_preserves_pre_section_content() -> None:
    report = _compliance_report(ok_urls=[], fail_urls=["https://example.com/fail"])
    report._text = f"=== Graph Audit Report ===\n\n--- Totals ---\n  Entities: 5\n\n{_COMPLIANCE_SECTION}\n"
    rendered = graph_cmd._render_audit_text(report, show_url_violations=True)
    assert "--- Totals ---" in rendered
    assert "Entities: 5" in rendered
    assert "https://example.com/fail" in rendered


def test_render_audit_text_includes_error_messages() -> None:
    url_result = _make_url_result("https://example.com/page", errors=["Missing schema:name"])
    report = _FakeReport(
        text=f"=== Graph Audit Report ===\n\n{_COMPLIANCE_SECTION}\n",
        url_results=[url_result],
    )
    rendered = graph_cmd._render_audit_text(report, show_url_violations=True)
    assert "Missing schema:name" in rendered
    assert "ERROR:" in rendered


def test_render_audit_text_deduplicates_repeated_errors() -> None:
    url_result = SimpleNamespace(
        url="https://example.com/page",
        errors=[
            SimpleNamespace(message="Value does not have class schema:Question", focus_node=f"urn:focus{i}")
            for i in range(3)
        ],
        warnings=[],
        error_count=3,
        warning_count=0,
        google_merchant=SimpleNamespace(eligible=False),
    )
    report = _FakeReport(
        text=f"=== Graph Audit Report ===\n\n{_COMPLIANCE_SECTION}\n",
        url_results=[url_result],
    )
    rendered = graph_cmd._render_audit_text(report, show_url_violations=True)
    assert rendered.count("Value does not have class schema:Question") == 1
    assert "×3" in rendered


def test_render_audit_text_single_error_shows_focus_node() -> None:
    url_result = SimpleNamespace(
        url="https://example.com/page",
        errors=[SimpleNamespace(message="Missing name", focus_node="urn:focus1")],
        warnings=[],
        error_count=1,
        warning_count=0,
        google_merchant=SimpleNamespace(eligible=False),
    )
    report = _FakeReport(
        text=f"=== Graph Audit Report ===\n\n{_COMPLIANCE_SECTION}\n",
        url_results=[url_result],
    )
    rendered = graph_cmd._render_audit_text(report, show_url_violations=True)
    assert "focus=urn:focus1" in rendered
    assert "×" not in rendered


def test_render_audit_text_no_marker_returns_full_text() -> None:
    report = _FakeReport(text="No compliance section here.", url_results=[])
    assert graph_cmd._render_audit_text(report) == "No compliance section here."


# ---------------------------------------------------------------------------
# Basic invocation
# ---------------------------------------------------------------------------


def test_audit_text_output_exit_0(monkeypatch) -> None:
    report = _FakeReport(text="Audit OK\n")
    _patch_sdk(monkeypatch, report)

    result = CliRunner().invoke(graph_cmd.app, ["audit", "graph.ttl"])

    assert result.exit_code == 0
    assert "Audit OK" in result.output


def test_audit_json_output_exit_0(monkeypatch) -> None:
    report = _FakeReport(data={"kpis": []})
    _patch_sdk(monkeypatch, report)

    result = CliRunner().invoke(graph_cmd.app, ["audit", "graph.ttl", "--format", "json"])

    assert result.exit_code == 0
    payload = json.loads(result.output)
    assert payload == {"kpis": []}


def test_audit_exit_1_on_load_errors(monkeypatch) -> None:
    report = _FakeReport(load_errors=["parse error"], text="FAIL\n")
    _patch_sdk(monkeypatch, report)

    result = CliRunner().invoke(graph_cmd.app, ["audit", "graph.ttl"])

    assert result.exit_code == 1
    assert "FAIL" in result.output


# ---------------------------------------------------------------------------
# Option forwarding
# ---------------------------------------------------------------------------


def test_audit_options_forwarded(monkeypatch) -> None:
    seen: dict = {}

    class _CapturingAuditor:
        def audit(self, path, options):
            seen["path"] = path
            seen["options"] = options
            return _FakeReport()

    class _AuditorCls:
        def __call__(self):
            return _CapturingAuditor()

    monkeypatch.setattr(graph_cmd, "_GraphAuditor", _AuditorCls())
    monkeypatch.setattr(graph_cmd, "_AuditOptions", lambda **kw: SimpleNamespace(**kw))

    result = CliRunner().invoke(
        graph_cmd.app,
        [
            "audit",
            "my_graph.ttl",
            "--depth",
            "3",
            "--list-isolated-iris",
            "--rich-snippets-granularity",
            "entities",
            "--workers",
            "4",
        ],
    )

    assert result.exit_code == 0
    assert seen["path"] == "my_graph.ttl"
    opts = seen["options"]
    assert opts.subgraph_depth == 3
    assert opts.list_isolated_components is True
    assert opts.rich_snippets_granularity == "entities"
    assert opts.max_workers == 4
    assert opts.profile is None


def test_audit_shape_options_forwarded(monkeypatch) -> None:
    seen: dict = {}

    class _Cap:
        def audit(self, path, options):
            seen["options"] = options
            return _FakeReport()

    class _Cls:
        def __call__(self):
            return _Cap()

    monkeypatch.setattr(graph_cmd, "_GraphAuditor", _Cls())
    monkeypatch.setattr(graph_cmd, "_AuditOptions", lambda **kw: SimpleNamespace(**kw))

    result = CliRunner().invoke(
        graph_cmd.app,
        [
            "audit",
            "graph.ttl",
            "--builtin-shape",
            "google-article.ttl",
            "--builtin-shape",
            "google-faq.ttl",
            "--shape",
            "./custom.ttl",
            "--issue-level",
            "error",
        ],
    )

    assert result.exit_code == 0
    opts = seen["options"]
    assert opts.builtin_shapes == ["google-article.ttl", "google-faq.ttl"]
    assert opts.exclude_builtin_shapes is None
    assert opts.extra_shapes == ["./custom.ttl"]
    assert opts.issue_level == "error"


def test_audit_exclude_builtin_shapes_forwarded(monkeypatch) -> None:
    seen: dict = {}

    class _Cap:
        def audit(self, path, options):
            seen["options"] = options
            return _FakeReport()

    class _Cls:
        def __call__(self):
            return _Cap()

    monkeypatch.setattr(graph_cmd, "_GraphAuditor", _Cls())
    monkeypatch.setattr(graph_cmd, "_AuditOptions", lambda **kw: SimpleNamespace(**kw))

    result = CliRunner().invoke(
        graph_cmd.app,
        ["audit", "graph.ttl", "--exclude-builtin-shape", "google-article.ttl"],
    )

    assert result.exit_code == 0
    opts = seen["options"]
    assert opts.builtin_shapes is None
    assert opts.exclude_builtin_shapes == ["google-article.ttl"]


def test_audit_builtin_and_exclude_are_mutually_exclusive(monkeypatch) -> None:
    _patch_sdk(monkeypatch)

    result = CliRunner().invoke(
        graph_cmd.app,
        [
            "audit",
            "graph.ttl",
            "--builtin-shape",
            "a.ttl",
            "--exclude-builtin-shape",
            "b.ttl",
        ],
    )

    assert result.exit_code == 1
    assert isinstance(result.exception, UsageError)
    assert "mutually exclusive" in str(result.exception)


def test_audit_no_shapes_passes_none(monkeypatch) -> None:
    seen: dict = {}

    class _Cap:
        def audit(self, path, options):
            seen["options"] = options
            return _FakeReport()

    class _Cls:
        def __call__(self):
            return _Cap()

    monkeypatch.setattr(graph_cmd, "_GraphAuditor", _Cls())
    monkeypatch.setattr(graph_cmd, "_AuditOptions", lambda **kw: SimpleNamespace(**kw))

    result = CliRunner().invoke(graph_cmd.app, ["audit", "graph.ttl"])

    assert result.exit_code == 0
    opts = seen["options"]
    assert opts.builtin_shapes is None
    assert opts.exclude_builtin_shapes is None
    assert opts.extra_shapes is None
    assert opts.issue_level == "warning"


# ---------------------------------------------------------------------------
# Profile resolution — global switch
# ---------------------------------------------------------------------------


def test_audit_no_profile_passes_none(monkeypatch) -> None:
    seen: dict = {}

    class _Cap:
        def audit(self, path, options):
            seen["profile"] = options.profile
            return _FakeReport()

    class _Cls:
        def __call__(self):
            return _Cap()

    monkeypatch.setattr(graph_cmd, "_GraphAuditor", _Cls())
    monkeypatch.setattr(graph_cmd, "_AuditOptions", lambda **kw: SimpleNamespace(**kw))

    result = CliRunner().invoke(graph_cmd.app, ["audit", "graph.ttl"], obj={"profile": None})

    assert result.exit_code == 0
    assert seen["profile"] is None


def test_audit_profile_loaded_from_global_switch(monkeypatch) -> None:
    fake_profile = SimpleNamespace(api_key="wl_test")
    seen: dict = {}

    monkeypatch.setattr(
        graph_cmd,
        "load_profile",
        lambda ctx: (fake_profile, "acme", Path("worai.toml")),
    )

    class _Cap:
        def audit(self, path, options):
            seen["profile"] = options.profile
            return _FakeReport()

    class _Cls:
        def __call__(self):
            return _Cap()

    monkeypatch.setattr(graph_cmd, "_GraphAuditor", _Cls())
    monkeypatch.setattr(graph_cmd, "_AuditOptions", lambda **kw: SimpleNamespace(**kw))

    result = CliRunner().invoke(graph_cmd.app, ["audit", "graph.ttl"], obj={"profile": "acme"})

    assert result.exit_code == 0
    assert seen["profile"] is fake_profile


def test_audit_profile_load_error_raises_usage_error(monkeypatch) -> None:
    monkeypatch.setattr(
        graph_cmd,
        "load_profile",
        lambda ctx: (_ for _ in ()).throw(ValueError("missing profile")),
    )
    monkeypatch.setattr(graph_cmd, "_GraphAuditor", object())
    monkeypatch.setattr(graph_cmd, "_AuditOptions", SimpleNamespace)

    result = CliRunner().invoke(graph_cmd.app, ["audit", "graph.ttl"], obj={"profile": "missing"})

    assert result.exit_code == 1
    assert isinstance(result.exception, UsageError)
    assert "missing profile" in str(result.exception)


# ---------------------------------------------------------------------------
# SDK unavailable
# ---------------------------------------------------------------------------


def test_audit_raises_usage_error_when_sdk_unavailable(monkeypatch) -> None:
    monkeypatch.setattr(graph_cmd, "_GraphAuditor", None)
    monkeypatch.setattr(graph_cmd, "_AuditOptions", None)

    result = CliRunner().invoke(graph_cmd.app, ["audit", "graph.ttl"])

    assert result.exit_code == 1
    assert isinstance(result.exception, UsageError)
    assert "wordlift-sdk>=6.11.0" in str(result.exception)


# ---------------------------------------------------------------------------
# Validation helpers
# ---------------------------------------------------------------------------


def test_normalize_rich_snippets_granularity() -> None:
    assert graph_cmd._normalize_rich_snippets_granularity("counts") == "counts"
    assert graph_cmd._normalize_rich_snippets_granularity("Entities") == "entities"
    with pytest.raises(UsageError, match="counts.*entities"):
        graph_cmd._normalize_rich_snippets_granularity("bad")


def test_audit_invalid_format(monkeypatch) -> None:
    _patch_sdk(monkeypatch)

    result = CliRunner().invoke(graph_cmd.app, ["audit", "graph.ttl", "--format", "xml"])

    assert result.exit_code == 1
    assert isinstance(result.exception, UsageError)


def test_audit_invalid_rich_snippets_granularity(monkeypatch) -> None:
    _patch_sdk(monkeypatch)

    result = CliRunner().invoke(
        graph_cmd.app, ["audit", "graph.ttl", "--rich-snippets-granularity", "bad"]
    )

    assert result.exit_code == 1
    assert isinstance(result.exception, UsageError)


# ---------------------------------------------------------------------------
# Auditor exception wrapping
# ---------------------------------------------------------------------------


def test_audit_unknown_builtin_shape_lists_available(monkeypatch) -> None:
    class _BrokenAuditor:
        def audit(self, path, options):
            raise ValueError("Unknown excluded builtin shape value: merchant")

    class _Cls:
        def __call__(self):
            return _BrokenAuditor()

    monkeypatch.setattr(graph_cmd, "_GraphAuditor", _Cls())
    monkeypatch.setattr(graph_cmd, "_AuditOptions", lambda **kw: SimpleNamespace(**kw))
    monkeypatch.setattr(
        graph_cmd,
        "_shape_specs_for_profile",
        lambda _p: ["google-merchant-listing.ttl", "google-article.ttl"],
    )

    result = CliRunner().invoke(
        graph_cmd.app, ["audit", "graph.ttl", "--exclude-builtin-shape", "merchant"]
    )

    assert result.exit_code == 1
    assert isinstance(result.exception, UsageError)
    error_msg = str(result.exception)
    assert "Unknown excluded builtin shape value: merchant" in error_msg
    assert "google-merchant-listing.ttl" in error_msg
    assert "google-article.ttl" in error_msg


def test_audit_wraps_auditor_exception(monkeypatch) -> None:
    class _BrokenAuditor:
        def audit(self, path, options):
            raise RuntimeError("parse failed")

    class _Cls:
        def __call__(self):
            return _BrokenAuditor()

    monkeypatch.setattr(graph_cmd, "_GraphAuditor", _Cls())
    monkeypatch.setattr(graph_cmd, "_AuditOptions", lambda **kw: SimpleNamespace(**kw))

    result = CliRunner().invoke(graph_cmd.app, ["audit", "graph.ttl"])

    assert result.exit_code == 1
    assert isinstance(result.exception, UsageError)
    assert "parse failed" in str(result.exception)



"""Materialization pipeline for YARRRML -> JSON-LD."""

from __future__ import annotations

from pathlib import Path

from wordlift_sdk.structured_data import YarrrmlPipeline


class MaterializationPipeline:
    """Normalizes mappings, materializes JSON-LD, and post-processes output."""

    def __init__(self, pipeline: YarrrmlPipeline | None = None) -> None:
        self._pipeline = pipeline or YarrrmlPipeline()

    def normalize(
        self, yarrrml: str, url: str, xhtml_path: Path, target_type: str | None
    ) -> tuple[str, list[dict]]:
        return self._pipeline.normalize_mappings(yarrrml, url, xhtml_path, target_type=target_type)

    def materialize(
        self, normalized_yarrrml: str, xhtml_path: Path, workdir: Path, url: str
    ) -> dict:
        return self._pipeline.materialize_jsonld(normalized_yarrrml, xhtml_path, workdir, url=url)

    def postprocess(
        self,
        jsonld_raw: dict,
        mappings: list[dict],
        cleaned_xhtml: str,
        dataset_uri: str,
        url: str,
        target_type: str | None,
    ) -> dict:
        return self._pipeline.postprocess_jsonld(
            jsonld_raw,
            mappings,
            cleaned_xhtml,
            dataset_uri,
            url,
            target_type=target_type,
        )

    def run(
        self,
        yarrrml: str,
        url: str,
        cleaned_xhtml: str,
        dataset_uri: str,
        xhtml_path: Path,
        workdir: Path,
        target_type: str | None,
    ) -> tuple[dict, list[dict]]:
        normalized_yarrrml, mappings = self.normalize(
            yarrrml, url, xhtml_path, target_type=target_type
        )
        jsonld_raw = self.materialize(normalized_yarrrml, xhtml_path, workdir, url=url)
        jsonld = self.postprocess(
            jsonld_raw,
            mappings,
            cleaned_xhtml,
            dataset_uri,
            url,
            target_type=target_type,
        )
        return jsonld, mappings



from __future__ import annotations

from pathlib import Path

import click
import pytest
from typer.testing import CliRunner

from worai.commands import (
    canonicalize_duplicate_pages as canonicalize_cmd,
    dedupe as dedupe_cmd,
    delete_entities_from_csv as delete_cmd,
    dil_import as dil_import_cmd,
    find_missing_names as find_missing_cmd,
    find_url_by_type as find_url_cmd,
    list_entities_outside_dataset as list_outside_cmd,
    patch as patch_cmd,
    seocheck as seocheck_cmd,
)
from worai.errors import UsageError


def test_seocheck_builds_argv_and_exits(monkeypatch: pytest.MonkeyPatch) -> None:
    seen: dict[str, object] = {}

    def _stub(argv: list[str]) -> int:
        seen["argv"] = argv
        return 7

    monkeypatch.setattr(seocheck_cmd.seocheck_cli, "run", _stub)
    with pytest.raises(click.exceptions.Exit) as exc:
        seocheck_cmd.run(
            sitemap_url="https://example.com/sitemap.xml",
            max_urls=2,
            output="out.json",
            output_summary="sum.json",
            save_html=True,
            no_report_ui=True,
            checks="status,canonical",
            disable_checks="ttfb",
            recheck_failed=True,
            recheck_from="report.json",
        )
    assert exc.value.exit_code == 7
    argv = seen["argv"]
    assert isinstance(argv, list)
    assert "--max-urls" in argv
    assert "--save-html" in argv
    assert "--recheck-failed" in argv


def test_find_missing_names_success_and_empty(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(find_missing_cmd, "find_run", lambda _p: ["https://example.com/a"])
    result = CliRunner().invoke(find_missing_cmd.app, ["data.ttl"])
    assert result.exit_code == 0
    assert "Found URLs for 1" in result.output

    monkeypatch.setattr(find_missing_cmd, "find_run", lambda _p: [])
    result = CliRunner().invoke(find_missing_cmd.app, ["data.ttl"])
    assert result.exit_code == 0
    assert "No schema:CollectionPage entities" in result.output


def test_find_missing_names_error_paths(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(find_missing_cmd, "find_run", lambda _p: (_ for _ in ()).throw(FileNotFoundError()))
    result = CliRunner().invoke(find_missing_cmd.app, ["missing.ttl"])
    assert result.exit_code == 1
    assert "was not found" in result.output

    monkeypatch.setattr(find_missing_cmd, "find_run", lambda _p: (_ for _ in ()).throw(RuntimeError("boom")))
    result = CliRunner().invoke(find_missing_cmd.app, ["bad.ttl"])
    assert result.exit_code == 1
    assert "An error occurred while parsing the file" in result.output


def test_find_url_by_type_paths(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(find_url_cmd, "resolve_types", lambda _types: [])
    result = CliRunner().invoke(find_url_cmd.app, ["x.ttl", "schema:Thing"])
    assert result.exit_code == 1
    assert "No valid schema types" in result.output

    monkeypatch.setattr(find_url_cmd, "resolve_types", lambda _types: ["Thing"])
    monkeypatch.setattr(find_url_cmd, "find_urls", lambda _opts: [("s1", "u1")])
    result = CliRunner().invoke(find_url_cmd.app, ["x.ttl", "schema:Thing"])
    assert result.exit_code == 0
    assert "u1" in result.output

    result = CliRunner().invoke(find_url_cmd.app, ["--show-id", "x.ttl", "schema:Thing"])
    assert result.exit_code == 0
    assert "s1" in result.output


def test_find_url_by_type_error_paths(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(find_url_cmd, "resolve_types", lambda _types: ["Thing"])
    monkeypatch.setattr(find_url_cmd, "find_urls", lambda _opts: (_ for _ in ()).throw(FileNotFoundError()))
    result = CliRunner().invoke(find_url_cmd.app, ["missing.ttl", "schema:Thing"])
    assert result.exit_code == 1
    assert "File not found" in result.output

    monkeypatch.setattr(find_url_cmd, "find_urls", lambda _opts: (_ for _ in ()).throw(RuntimeError("boom")))
    result = CliRunner().invoke(find_url_cmd.app, ["bad.ttl", "schema:Thing"])
    assert result.exit_code == 1
    assert "Error parsing graph file" in result.output


def test_dil_import_verify_csv_and_run(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    csv_path = tmp_path / "dils.csv"
    csv_path.write_text("source_url,target_url,target_anchor_text,target_position\ns,t,a,1\n", encoding="utf-8")
    seen: dict[str, object] = {}
    monkeypatch.setattr(dil_import_cmd, "import_run", lambda options: seen.setdefault("options", options))
    result = CliRunner().invoke(dil_import_cmd.app, ["wl_1", str(csv_path)])
    assert result.exit_code == 0
    assert seen["options"].api_key == "wl_1"


def test_dil_import_verify_csv_errors(tmp_path: Path) -> None:
    with pytest.raises(Exception):
        dil_import_cmd.verify_csv(str(tmp_path / "missing.csv"))

    txt_path = tmp_path / "x.txt"
    txt_path.write_text("x", encoding="utf-8")
    with pytest.raises(Exception):
        dil_import_cmd.verify_csv(str(txt_path))


def test_sdk_key_wrappers_use_profile_loader(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    csv_path = tmp_path / "in.csv"
    csv_path.write_text("iri\nx\n", encoding="utf-8")

    monkeypatch.setattr(canonicalize_cmd, "resolve_profile_api_key", lambda _ctx: ("wl", "default", Path("w.toml")))
    monkeypatch.setattr(dedupe_cmd, "resolve_profile_api_key", lambda _ctx: ("wl", "default", Path("w.toml")))
    monkeypatch.setattr(delete_cmd, "resolve_profile_api_key", lambda _ctx: ("wl", "default", Path("w.toml")))
    monkeypatch.setattr(list_outside_cmd, "resolve_profile_api_key", lambda _ctx: ("wl", "default", Path("w.toml")))
    monkeypatch.setattr(patch_cmd, "resolve_profile_api_key", lambda _ctx: ("wl", "default", Path("w.toml")))

    monkeypatch.setattr(canonicalize_cmd, "canonicalize_run", lambda _opts: None)
    monkeypatch.setattr(dedupe_cmd, "dedupe_run", lambda _opts: None)
    monkeypatch.setattr(delete_cmd, "delete_entities", lambda _opts: None)
    monkeypatch.setattr(list_outside_cmd, "run", lambda _opts: None)
    monkeypatch.setattr(patch_cmd, "patch_run", lambda _opts: None)

    canonicalize_cmd.run(None, "in.csv", "out.csv", None, None, "e", 10, "28d", "clicks")
    dedupe_cmd.run(None, "e", False, 0.0, True)
    delete_cmd.run(None, str(csv_path), 10)
    list_outside_cmd.main(None, None, "e", "b", 0, "q")
    patch_cmd.run(None, "f.ttl", False, False, False, None, 1, 1)


def test_sdk_key_wrappers_raise_without_key(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    monkeypatch.setattr(canonicalize_cmd, "resolve_profile_api_key", lambda _ctx: (None, "default", Path("w.toml")))
    with pytest.raises(UsageError):
        canonicalize_cmd.run(None, "in.csv", "out.csv", None, None, "e", 10, "28d", "clicks")



"""Structured-data inventory workflows backed by SDK ingestion APIs."""

from __future__ import annotations

import csv
import json
import tempfile
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable
from urllib.parse import quote, urlparse

from google.auth.transport.requests import AuthorizedSession

from worai.core.google_oauth import SHEETS_SCOPES, load_google_credentials
from worai.core.ingest import DEFAULT_INGEST_LOADER, resolve_ingest_settings
from worai.core.url_sources.file import load_urls_from_file
from worai.core.url_sources.models import UrlSourceOptions
from worai.core.url_sources.resolver import resolve_url_records
from worai.core.url_sources.sheets import is_spreadsheet_source
from worai.errors import UsageError

DEFAULT_BASE_URL = "https://api.wordlift.io"
DEFAULT_TIMEOUT = 30.0


@dataclass
class InventoryRow:
    url: str
    faq_markup: str
    faq_markup_from_graph: str
    types: str
    structured_data: str


@dataclass
class InventoryOptions:
    source: str
    api_key: str
    base_url: str = DEFAULT_BASE_URL
    timeout: float = DEFAULT_TIMEOUT
    source_sheet_name: str | None = None
    url_regex: str | None = None
    output: str | None = None
    destination_spreadsheet_id: str | None = None
    destination_sheet_name: str | None = None
    client_secrets: str | None = None
    token: str = "oauth_token.json"
    port: int = 8080
    concurrency: str = "auto"
    ingest_source: str | None = None
    ingest_loader: str | None = None
    ingest_passthrough_when_html: bool | None = None
    ingest_timeout_ms: int | None = None
    ingest_retry_attempts: int | None = None
    ingest_retry_backoff_ms: int | None = None
    playwright_wait_until: str | None = None
    source_type: str | None = None


def _headers() -> list[str]:
    return ["url", "faq_markup", "faq_markup_from_graph", "types", "structured_data"]


def _normalize_loader(value: str) -> str:
    if value == "auto":
        return DEFAULT_INGEST_LOADER
    return value


def _infer_source_mode(source: str, sheet_name: str | None) -> str:
    if is_spreadsheet_source(source):
        return "sheets"

    path = Path(source)
    if path.exists():
        suffix = path.suffix.lower()
        if suffix in {".ttl", ".json"}:
            return "local"
        if suffix in {".xml", ".gz"}:
            return "sitemap"
        return "urls"

    parsed = urlparse(source)
    if parsed.scheme in {"http", "https", "file"}:
        return "sitemap"

    if sheet_name:
        return "sheets"

    raise UsageError(
        "Cannot infer ingest source from SOURCE. Pass --ingest-source explicitly "
        "(urls|sitemap|sheets|local)."
    )


def _build_source_bundle(options: InventoryOptions) -> dict[str, object]:
    resolved = resolve_ingest_settings(
        context="structured_data_inventory",
        new_source=options.ingest_source,
        new_loader=options.ingest_loader,
        new_passthrough_when_html=options.ingest_passthrough_when_html,
        legacy_source=options.source_type,
    )

    source_name = (
        _infer_source_mode(options.source, options.source_sheet_name)
        if resolved.source == "auto"
        else resolved.source
    )

    bundle: dict[str, object] = {
        "INGEST_SOURCE": source_name,
        "INGEST_LOADER": _normalize_loader(resolved.loader),
        "INGEST_PASSTHROUGH_WHEN_HTML": resolved.passthrough_when_html,
    }

    if options.url_regex:
        bundle["URL_REGEX"] = options.url_regex
    if options.ingest_timeout_ms is not None:
        bundle["INGEST_TIMEOUT_MS"] = options.ingest_timeout_ms
    if options.ingest_retry_attempts is not None:
        bundle["INGEST_RETRY_ATTEMPTS"] = options.ingest_retry_attempts
    if options.ingest_retry_backoff_ms is not None:
        bundle["INGEST_RETRY_BACKOFF_MS"] = options.ingest_retry_backoff_ms
    if options.playwright_wait_until:
        bundle["PLAYWRIGHT_WAIT_UNTIL"] = options.playwright_wait_until

    if source_name == "urls":
        source_path = Path(options.source)
        if not source_path.exists():
            raise UsageError("INGEST_SOURCE=urls requires a local URL list file source.")
        bundle["URLS"] = load_urls_from_file(source_path)
        return bundle

    if source_name == "sitemap":
        bundle["SITEMAP_URL"] = options.source
        return bundle

    if source_name == "sheets":
        if not options.source_sheet_name:
            raise UsageError("When source is a Google Spreadsheet, --sheet-name is required.")

        # Keep existing OAuth-based sheets UX in worai, then pass resolved URLs to SDK.
        records = resolve_url_records(
            UrlSourceOptions(
                source=options.source,
                timeout=options.timeout,
                sheet_name=options.source_sheet_name,
                url_regex=options.url_regex,
                client_secrets=options.client_secrets,
                token=options.token,
                port=options.port,
                ingest_source="sheets",
            )
        )
        bundle["INGEST_SOURCE"] = "urls"
        bundle.pop("URL_REGEX", None)
        bundle["URLS"] = [record.url for record in records]
        return bundle

    if source_name == "local":
        source_path = Path(options.source)
        if source_path.is_file() and source_path.suffix.lower() == ".json":
            bundle["INGEST_LOCAL_FILE"] = options.source
            return bundle

        # Preserve debug-cloud/local compatibility by passing resolved local items.
        records = resolve_url_records(
            UrlSourceOptions(
                source=options.source,
                timeout=options.timeout,
                sheet_name=options.source_sheet_name,
                url_regex=options.url_regex,
                client_secrets=options.client_secrets,
                token=options.token,
                port=options.port,
                ingest_source="local",
                source_type=options.source_type,
            )
        )
        items: list[dict[str, Any]] = []
        for index, record in enumerate(records):
            item: dict[str, Any] = {
                "id": f"local:{index}",
                "url": record.url,
            }
            if record.html is not None:
                item["html"] = record.html
            if record.metadata:
                item["metadata"] = record.metadata
            items.append(item)
        bundle["INGEST_LOCAL_ITEMS"] = items
        return bundle

    raise UsageError(f"Unsupported ingest source: {source_name}")


def _normalize_yes_no(value: Any) -> str:
    if isinstance(value, bool):
        return "yes" if value else "no"
    token = str(value or "").strip().lower()
    if token in {"yes", "true", "1"}:
        return "yes"
    if token in {"no", "false", "0", ""}:
        return "no"
    return token


def _normalize_structured_data(value: Any) -> str:
    if value is None:
        return ""
    if isinstance(value, str):
        return value
    return json.dumps(value, ensure_ascii=False, separators=(",", ":"))


def _normalize_types(value: Any) -> str:
    if value is None:
        return ""
    if isinstance(value, str):
        return value
    if isinstance(value, list):
        return ",".join(str(item) for item in value if str(item).strip())
    return str(value)


def _record_to_row(record: dict[str, Any]) -> InventoryRow:
    return InventoryRow(
        url=str(record.get("url", "")),
        faq_markup=_normalize_yes_no(record.get("faq_markup")),
        faq_markup_from_graph=_normalize_yes_no(record.get("faq_markup_from_graph")),
        types=_normalize_types(record.get("types")),
        structured_data=_normalize_structured_data(record.get("structured_data")),
    )


def _rows_from_csv(path: str) -> list[InventoryRow]:
    with open(path, "r", encoding="utf-8", newline="") as handle:
        reader = csv.DictReader(handle)
        return [_record_to_row(row) for row in reader]


def build_inventory(
    options: InventoryOptions,
    *,
    on_progress: Callable[[dict[str, Any]], None] | None = None,
) -> list[InventoryRow]:
    try:
        from wordlift_sdk.ingestion import create_structured_data_inventory_from_ingestion
    except Exception as exc:  # pragma: no cover - depends on installed SDK
        raise UsageError(
            "create_structured_data_inventory_from_ingestion requires wordlift-sdk>=6.9.0."
        ) from exc

    source_bundle = _build_source_bundle(options)

    if options.output:
        output_csv = options.output
        cleanup_path: str | None = None
    else:
        tmp_file = tempfile.NamedTemporaryFile(prefix="worai-inventory-", suffix=".csv", delete=False)
        output_csv = tmp_file.name
        tmp_file.close()
        cleanup_path = output_csv

    try:
        result = create_structured_data_inventory_from_ingestion(
            source_bundle=source_bundle,
            api_key=options.api_key,
            base_url=options.base_url,
            output_csv=output_csv,
            on_progress=on_progress,
        )

        if hasattr(result, "to_dict"):
            records = result.to_dict(orient="records")
            if isinstance(records, list):
                return [_record_to_row(record) for record in records if isinstance(record, dict)]

        return _rows_from_csv(output_csv)
    except UsageError:
        raise
    except Exception as exc:
        detail = str(exc).strip()
        detail_suffix = (
            f" Details: {exc.__class__.__name__}: {detail}"
            if detail
            else f" Details: {exc.__class__.__name__}."
        )
        raise UsageError(
            "Failed to create structured-data inventory from ingestion source. "
            "Check source access, ingest settings, and SDK compatibility."
            f"{detail_suffix}"
        ) from exc
    finally:
        if cleanup_path:
            try:
                Path(cleanup_path).unlink(missing_ok=True)
            except Exception:
                pass


def write_csv(path: str, rows: list[InventoryRow]) -> None:
    with open(path, "w", newline="", encoding="utf-8") as handle:
        writer = csv.DictWriter(handle, fieldnames=_headers())
        writer.writeheader()
        for row in rows:
            writer.writerow(row.__dict__)


def _get_sheet_titles(session: AuthorizedSession, spreadsheet_id: str) -> list[str]:
    url = f"https://sheets.googleapis.com/v4/spreadsheets/{spreadsheet_id}?fields=sheets.properties.title"
    response = session.get(url)
    if response.status_code != 200:
        raise RuntimeError(f"Failed to read spreadsheet metadata ({response.status_code}): {response.text}")
    data = response.json()
    sheets = data.get("sheets") or []
    titles: list[str] = []
    for sheet in sheets:
        title = ((sheet or {}).get("properties") or {}).get("title")
        if isinstance(title, str) and title:
            titles.append(title)
    return titles


def _ensure_sheet_exists(
    session: AuthorizedSession,
    spreadsheet_id: str,
    sheet_title: str,
) -> None:
    if sheet_title in _get_sheet_titles(session, spreadsheet_id):
        return
    url = f"https://sheets.googleapis.com/v4/spreadsheets/{spreadsheet_id}:batchUpdate"
    payload = {
        "requests": [
            {
                "addSheet": {
                    "properties": {"title": sheet_title},
                }
            }
        ]
    }
    response = session.post(url, json=payload)
    if response.status_code != 200:
        raise RuntimeError(
            f"Failed to create destination sheet '{sheet_title}' ({response.status_code}): {response.text}"
        )


def write_google_sheet(
    spreadsheet_id: str,
    sheet_name: str,
    rows: list[InventoryRow],
    *,
    client_secrets: str | None,
    token: str,
    port: int,
) -> None:
    creds = load_google_credentials(
        client_secrets_path=client_secrets,
        token_path=token,
        port=port,
        scopes=SHEETS_SCOPES,
    )
    session = AuthorizedSession(creds)

    _ensure_sheet_exists(session, spreadsheet_id, sheet_name)
    range_name = f"'{sheet_name}'!A1"
    encoded_range = quote(range_name, safe="")

    clear_url = (
        f"https://sheets.googleapis.com/v4/spreadsheets/{spreadsheet_id}/values/{encoded_range}:clear"
    )
    clear_resp = session.post(clear_url, json={})
    if clear_resp.status_code not in {200, 204}:
        raise RuntimeError(f"Failed to clear spreadsheet ({clear_resp.status_code}): {clear_resp.text}")

    values: list[list[str]] = [_headers()]
    for row in rows:
        values.append([
            row.url,
            row.faq_markup,
            row.faq_markup_from_graph,
            row.types,
            row.structured_data,
        ])

    update_url = (
        f"https://sheets.googleapis.com/v4/spreadsheets/{spreadsheet_id}/values/{encoded_range}"
        "?valueInputOption=RAW"
    )
    payload = {
        "range": range_name,
        "majorDimension": "ROWS",
        "values": values,
    }
    update_resp = session.put(update_url, json=payload)
    if update_resp.status_code != 200:
        raise RuntimeError(f"Failed to write spreadsheet ({update_resp.status_code}): {update_resp.text}")


def run_inventory(
    options: InventoryOptions,
    *,
    on_progress: Callable[[dict[str, Any]], None] | None = None,
) -> list[InventoryRow]:
    if bool(options.output) == bool(options.destination_spreadsheet_id):
        raise UsageError(
            "Provide exactly one destination: --output or --destination-sheet-id."
        )
    if options.destination_spreadsheet_id and not options.destination_sheet_name:
        raise UsageError(
            "--destination-sheet-name is required with --destination-sheet-id."
        )

    rows = build_inventory(options, on_progress=on_progress)

    if options.output:
        write_csv(options.output, rows)
    if options.destination_spreadsheet_id:
        write_google_sheet(
            options.destination_spreadsheet_id,
            options.destination_sheet_name or "",
            rows,
            client_secrets=options.client_secrets,
            token=options.token,
            port=options.port,
        )

    return rows



from __future__ import annotations

from pathlib import Path

import pytest

from worai.core.url_sources.models import UrlRecord, UrlSourceOptions
from worai.core.url_sources import resolver as mod
from worai.errors import UsageError


def test_dedupe_normalize_and_http_predicate() -> None:
    assert mod._dedupe([" a ", "", "a", "b"]) == ["a", "b"]
    assert mod._is_http_url("https://e.com") is True
    assert mod._is_http_url("ftp://e.com") is False

    rows = mod._normalize_records([
        UrlRecord(url=" "),
        UrlRecord(url="ftp://e.com"),
        UrlRecord(url="https://e.com", html="h"),
        UrlRecord(url="https://e.com"),
    ])
    assert len(rows) == 1
    assert rows[0].html == "h"

    compiled = mod._compile_url_regex(r"ok")
    assert compiled is not None and compiled.search("https://e.com/ok")
    assert mod._compile_url_regex(None) is None
    assert mod._compile_url_regex("  ") is None


def test_resolve_url_records_explicit_source_modes(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    monkeypatch.setattr(mod, "resolve_ingest_settings", lambda **_k: type("I", (), {"source": "urls"})())
    with pytest.raises(UsageError, match="requires a local URL list file"):
        mod.resolve_url_records(UrlSourceOptions(source=str(tmp_path / "missing.txt"), ingest_source="urls"))

    p = tmp_path / "u.txt"
    p.write_text("https://e.com/a\n", encoding="utf-8")
    monkeypatch.setattr(mod, "load_urls_from_file", lambda _p: ["https://e.com/a"])
    rows = mod.resolve_url_records(UrlSourceOptions(source=str(p), ingest_source="urls"))
    assert rows[0].source_type == "file"

    monkeypatch.setattr(mod, "resolve_ingest_settings", lambda **_k: type("I", (), {"source": "sitemap"})())
    monkeypatch.setattr(mod, "load_urls_from_sitemap", lambda *_a, **_k: ["https://e.com/s"])
    rows = mod.resolve_url_records(UrlSourceOptions(source="https://e.com/sitemap.xml", ingest_source="sitemap"))
    assert rows[0].source_type == "sitemap"

    monkeypatch.setattr(mod, "resolve_ingest_settings", lambda **_k: type("I", (), {"source": "sheets"})())
    with pytest.raises(UsageError, match="requires a Google Spreadsheet"):
        mod.resolve_url_records(UrlSourceOptions(source="https://e.com/sitemap.xml", ingest_source="sheets", sheet_name="S"))

    monkeypatch.setattr(mod, "is_spreadsheet_source", lambda _s: True)
    with pytest.raises(UsageError, match="sheet-name is required"):
        mod.resolve_url_records(UrlSourceOptions(source="sheetid", ingest_source="sheets"))

    monkeypatch.setattr(mod, "load_urls_from_sheet", lambda *_a, **_k: ["https://e.com/a"])
    rows = mod.resolve_url_records(UrlSourceOptions(source="sheetid", ingest_source="sheets", sheet_name="S"))
    assert rows[0].source_type == "sheet"


def test_resolve_url_records_path_and_fallback(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    monkeypatch.setattr(mod, "resolve_ingest_settings", lambda **_k: type("I", (), {"source": "auto"})())

    ttl = tmp_path / "x.ttl"
    ttl.write_text("", encoding="utf-8")
    monkeypatch.setattr(mod, "load_records_from_debug_cloud", lambda _s: [UrlRecord(url="https://e.com/d")])
    assert mod.resolve_url_records(UrlSourceOptions(source=str(ttl)))[0].url == "https://e.com/d"

    xml = tmp_path / "x.xml"
    xml.write_text("", encoding="utf-8")
    monkeypatch.setattr(mod, "load_urls_from_sitemap", lambda *_a, **_k: ["https://e.com/x"])
    assert mod.resolve_url_records(UrlSourceOptions(source=str(xml)))[0].source_type == "sitemap"

    txt = tmp_path / "x.txt"
    txt.write_text("", encoding="utf-8")
    monkeypatch.setattr(mod, "load_urls_from_file", lambda _p: ["https://e.com/f"])
    assert mod.resolve_url_records(UrlSourceOptions(source=str(txt)))[0].source_type == "file"

    rows = mod.resolve_url_records(UrlSourceOptions(source="file:///tmp/sitemap.xml"))
    assert rows[0].source_type == "sitemap"

    with pytest.raises(UsageError, match="SOURCE must be"):
        mod.resolve_url_records(UrlSourceOptions(source="not-a-source"))


def test_resolve_url_records_invalid_url_regex(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(mod, "resolve_ingest_settings", lambda **_k: type("I", (), {"source": "sitemap"})())
    monkeypatch.setattr(mod, "load_urls_from_sitemap", lambda *_a, **_k: ["https://e.com/a"])

    with pytest.raises(UsageError, match="Invalid --url-regex pattern"):
        mod.resolve_url_records(
            UrlSourceOptions(
                source="https://e.com/sitemap.xml",
                ingest_source="sitemap",
                url_regex="(",
            )
        )



"""Agent launcher command with MCP + skills bootstrap."""

from __future__ import annotations

import json
import os
import shlex
import shutil
import subprocess
import sys
import tempfile
from dataclasses import dataclass
from enum import Enum
from pathlib import Path

import click
import typer
from rich.console import Console as _RichConsole
from typer.core import TyperGroup

from worai.agent.mcp_server import run_stdio_server, start_http_server
from worai.agent.skills import wordlift_md_path, write_skill_file
from worai.config import resolve_config_path
from worai.errors import UsageError

class AgentGroup(TyperGroup):
    def invoke(self, ctx: click.Context) -> object:
        protected_args = list(ctx._protected_args or [])
        if protected_args and protected_args[0].startswith("-"):
            ctx.args = [*protected_args, *ctx.args]
            ctx._protected_args = []
        return super().invoke(ctx)


app = typer.Typer(
    add_completion=False,
    no_args_is_help=False,
    cls=AgentGroup,
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True},
)
mcp_app = typer.Typer(add_completion=False)
app.add_typer(mcp_app, name="mcp", help="Run worai MCP server utilities.")

_WORAI_SERVER_NAME = "worai"
_WORDLIFT_SERVER_NAME = "wordlift"
_DEFAULT_WORDLIFT_MCP_URL = "https://mcp.wordlift.io/sse"


class AgentCli(str, Enum):
    codex = "codex"
    claude = "claude"
    gemini = "gemini"


class McpTransport(str, Enum):
    stdio = "stdio"
    http = "http"


_WORDLIFT_ASCII_ART = (
    "██     ██  ██████  ██████  ██████  ██      ██ ███████ ████████\n"
    "██     ██ ██    ██ ██   ██ ██   ██ ██      ██ ██         ██   \n"
    "██  █  ██ ██    ██ ██████  ██   ██ ██      ██ █████      ██   \n"
    "██ ███ ██ ██    ██ ██   ██ ██   ██ ██      ██ ██         ██   \n"
    " ███ ███   ██████  ██   ██ ██████  ███████ ██ ██         ██   "
)

_BANNER_CONSOLE = _RichConsole(highlight=False)


def _print_wordlift_banner() -> None:
    _BANNER_CONSOLE.print(f"\n[bold #3452DB]{_WORDLIFT_ASCII_ART}[/]\n")
    _BANNER_CONSOLE.print("🚀 [bold]WordLift CLI[/bold] - Universal SEO Assistant")
    _BANNER_CONSOLE.rule(style="#3452DB")
    _BANNER_CONSOLE.print("🤖 [bold #3452DB]Agent WordLift[/] - SEO Specialist Mode Activated")
    _BANNER_CONSOLE.print(f"📁 [dim]Project: {Path.cwd()}[/dim]\n")


@dataclass(frozen=True)
class AgentLaunchAssets:
    temp_dir: Path
    skills_path: Path
    wordlift_md_path: Path
    mcp_config_path: Path
    gemini_home: Path


def _project_root() -> Path:
    # /.../worai/commands/agent.py -> /... (repo root for local dev)
    return Path(__file__).resolve().parents[2]


def _mcp_env() -> dict[str, str]:
    env = {"WORAI_DISABLE_UPDATE_CHECK": "1"}
    project_root = str(_project_root())
    existing_pythonpath = os.environ.get("PYTHONPATH", "").strip()
    if existing_pythonpath:
        parts = [project_root, *[part for part in existing_pythonpath.split(os.pathsep) if part]]
        deduped: list[str] = []
        for part in parts:
            if part not in deduped:
                deduped.append(part)
        env["PYTHONPATH"] = os.pathsep.join(deduped)
    else:
        env["PYTHONPATH"] = project_root
    return env


def _resolve_executable(agent_cli: AgentCli, override: str | None) -> str:
    if override and override.strip():
        return override.strip()
    return agent_cli.value


def _require_executable(name: str) -> None:
    if Path(name).is_absolute() and Path(name).exists():
        return
    if not shutil_which(name):
        raise UsageError(
            f"Executable not found: {name}. Install it or pass --agent-exec with the binary path."
        )


def shutil_which(name: str) -> str | None:
    from shutil import which

    return which(name)


def _build_mcp_command(config: str | None, profile: str | None) -> list[str]:
    cmd = [sys.executable, "-m", "worai.agent.mcp_server"]
    if config:
        cmd.extend(["--config", config])
    if profile:
        cmd.extend(["--profile", profile])
    return cmd


def _create_session_dir() -> Path:
    workspace_tmp = Path.cwd() / ".gemini" / "tmp"
    try:
        workspace_tmp.mkdir(parents=True, exist_ok=True)
        return Path(tempfile.mkdtemp(prefix="worai-agent-", dir=str(workspace_tmp)))
    except OSError:
        return Path(tempfile.mkdtemp(prefix="worai-agent-"))


def _wordlift_mcp_url() -> str:
    return os.environ.get("WORAI_WORDLIFT_MCP_URL", _DEFAULT_WORDLIFT_MCP_URL).strip()


def _wordlift_transport(url: str) -> str:
    lower = url.lower()
    if lower.endswith("/sse") or "/sse?" in lower:
        return "sse"
    return "http"


def _codex_transport(url: str) -> str:
    # Codex config supports streamable HTTP transport for URL-based servers.
    # Map SSE URLs to http to avoid stdio/url validation conflicts when an existing
    # `mcp_servers.<name>` entry is present from prior sessions.
    del url
    return "http"


def _configured_codex_mcp_servers(agent_exec: str) -> set[str]:
    completed = subprocess.run(
        [agent_exec, "mcp", "list", "--json"],
        check=False,
        capture_output=True,
        text=True,
    )
    if completed.returncode != 0:
        return set()
    try:
        payload = json.loads(completed.stdout or "[]")
    except json.JSONDecodeError:
        return set()
    if not isinstance(payload, list):
        return set()
    names: set[str] = set()
    for item in payload:
        if not isinstance(item, dict):
            continue
        name = item.get("name")
        if isinstance(name, str) and name.strip():
            names.add(name.strip())
    return names


def _gemini_env(gemini_home: Path) -> dict[str, str]:
    env = os.environ.copy()
    env["GEMINI_CLI_HOME"] = str(gemini_home)
    return env


def _setup_gemini_home(temp_dir: Path) -> Path:
    """Create an isolated GEMINI_CLI_HOME, seeding auth files from ~/.gemini/ (skipping settings)."""
    gemini_home = temp_dir / "gemini-home"
    gemini_dir = gemini_home / ".gemini"
    gemini_dir.mkdir(parents=True, exist_ok=True)
    source = Path.home() / ".gemini"
    _SKIP = {"settings.json", "tmp", "checkpoints", "history", "logs"}
    if source.is_dir():
        for item in source.iterdir():
            if item.name in _SKIP or item.name.startswith("."):
                continue
            try:
                if item.is_file():
                    shutil.copy2(item, gemini_dir / item.name)
            except OSError:
                pass
    return gemini_home


def _create_assets(
    *,
    config: str | None,
    profile: str | None,
    mcp_url: str | None = None,
) -> AgentLaunchAssets:
    temp_dir = _create_session_dir()
    gemini_home = _setup_gemini_home(temp_dir)
    gemini_dir = gemini_home / ".gemini"

    # Skills go in the default location Gemini expects: GEMINI_CLI_HOME/.gemini/skills/
    skills_dir = gemini_dir / "skills"
    skills_dir.mkdir(parents=True, exist_ok=True)
    skills_path = write_skill_file(skills_dir, project_root=Path.cwd())

    # WORDLIFT.md goes in GEMINI_CLI_HOME/.gemini/ so Gemini discovers it as a global context file.
    # contextFileName in settings tells Gemini to look for "WORDLIFT.md" by name.
    wl_md_dest = gemini_dir / "WORDLIFT.md"
    wl_md_dest.write_text(
        wordlift_md_path().read_text(encoding="utf-8")
        + "\n\n---\n\n"
        + skills_path.read_text(encoding="utf-8"),
        encoding="utf-8",
    )

    mcp_command = _build_mcp_command(config=config, profile=profile)
    mcp_config_path = temp_dir / "mcp.json"
    wordlift_url = _wordlift_mcp_url()
    if mcp_url:
        worai_server_entry: dict[str, object] = {"type": "http", "url": mcp_url}
    else:
        worai_server_entry = {
            "command": mcp_command[0],
            "args": mcp_command[1:],
            "env": _mcp_env(),
        }
    mcp_servers = {
        _WORAI_SERVER_NAME: worai_server_entry,
        _WORDLIFT_SERVER_NAME: {
            "type": _wordlift_transport(wordlift_url),
            "url": wordlift_url,
        },
    }
    (gemini_dir / "settings.json").write_text(
        json.dumps(
            {
                "ui": {"hideBanner": True},
                "mcp": {"allowed": [_WORAI_SERVER_NAME, _WORDLIFT_SERVER_NAME]},
                "mcpServers": mcp_servers,
                "experimental": {"skills": True},
                "contextFileName": "WORDLIFT.md",
                "context": {"fileName": "WORDLIFT.md"},
            },
            indent=2,
        ),
        encoding="utf-8",
    )
    # mcp.json for Claude's --mcp-config flag
    mcp_config_path.write_text(json.dumps({"mcpServers": mcp_servers}, indent=2), encoding="utf-8")
    return AgentLaunchAssets(
        temp_dir=temp_dir,
        skills_path=skills_path,
        wordlift_md_path=wl_md_dest,
        mcp_config_path=mcp_config_path,
        gemini_home=gemini_home,
    )


def _build_session_prompt(assets: AgentLaunchAssets) -> str:
    # Combined WORDLIFT.md (persona + branding + skill guidance) for Claude and Codex.
    # Gemini loads it silently via contextFileName in settings.json.
    return assets.wordlift_md_path.read_text(encoding="utf-8")


def _to_toml_inline_table(values: dict[str, str]) -> str:
    parts = [f"{key}={json.dumps(value)}" for key, value in values.items()]
    return "{" + ", ".join(parts) + "}"


def _build_launch_command(
    *,
    agent_cli: AgentCli,
    agent_exec: str,
    assets: AgentLaunchAssets,
    config: str | None,
    profile: str | None,
    transport: McpTransport,
    mcp_url: str | None,
    passthrough: list[str],
) -> tuple[list[str], dict[str, str]]:
    prompt = _build_session_prompt(assets)
    mcp_command = _build_mcp_command(config=config, profile=profile)
    if agent_cli == AgentCli.codex:
        wordlift_url = _wordlift_mcp_url()
        wordlift_transport = _codex_transport(wordlift_url)
        configured_servers = _configured_codex_mcp_servers(agent_exec)
        cmd = [agent_exec]
        if transport == McpTransport.http:
            if not mcp_url:
                raise UsageError("mcp_url is required for HTTP transport")
            if _WORAI_SERVER_NAME not in configured_servers:
                worai_table = (
                    "{"
                    f"url={json.dumps(mcp_url)}, "
                    'type="http", '
                    "enabled=true, "
                    "startup_timeout_sec=60"
                    "}"
                )
                cmd.extend(["-c", f"mcp_servers.{_WORAI_SERVER_NAME}={worai_table}"])
        else:
            mcp_env = _mcp_env()
            if _WORAI_SERVER_NAME not in configured_servers:
                worai_table = (
                    "{"
                    f"command={json.dumps(mcp_command[0])}, "
                    f"args={json.dumps(mcp_command[1:])}, "
                    f"env={_to_toml_inline_table(mcp_env)}, "
                    "enabled=true, "
                    "startup_timeout_sec=60"
                    "}"
                )
                cmd.extend(["-c", f"mcp_servers.{_WORAI_SERVER_NAME}={worai_table}"])
        if _WORDLIFT_SERVER_NAME not in configured_servers:
            cmd.extend(
                [
                    "-c",
                    (
                        f"mcp_servers.{_WORDLIFT_SERVER_NAME}="
                        "{"
                        f"url={json.dumps(wordlift_url)}, "
                        f"type={json.dumps(wordlift_transport)}, "
                        "enabled=true, "
                        "startup_timeout_sec=60"
                        "}"
                    ),
                ]
            )
        cmd.extend([*passthrough, prompt])
        return cmd, {}

    if agent_cli == AgentCli.claude:
        cmd = [
            agent_exec,
            "--mcp-config",
            str(assets.mcp_config_path),
            "--strict-mcp-config",
            "--append-system-prompt",
            prompt,
            *passthrough,
        ]
        return cmd, {}

    if agent_cli == AgentCli.gemini:
        cmd = [agent_exec, *passthrough]
        return cmd, {"WORAI_GEMINI_MCP_COMMAND": json.dumps(mcp_command)}

    raise UsageError(f"Unsupported agent CLI: {agent_cli.value}")



@app.callback(invoke_without_command=True)
def run(
    ctx: typer.Context,
    agent_cli: AgentCli = typer.Option(
        AgentCli.codex,
        "--agent-cli",
        "--cli",
        help="Agent CLI to launch: codex|claude|gemini.",
    ),
    agent_exec: str | None = typer.Option(
        None,
        "--agent-exec",
        help="Override agent CLI executable path.",
    ),
    config: str | None = typer.Option(
        None,
        "--config",
        help="Session default config path forwarded to MCP tools.",
    ),
    profile: str | None = typer.Option(
        None,
        "--profile",
        help="Session default profile forwarded to MCP tools.",
    ),
    mcp_transport: McpTransport = typer.Option(
        McpTransport.http,
        "--mcp-transport",
        help="MCP transport to expose for the agent session: stdio|http.",
    ),
    mcp_host: str = typer.Option(
        "127.0.0.1",
        "--mcp-host",
        help="Host for in-process MCP HTTP server.",
    ),
    mcp_port: int = typer.Option(
        0,
        "--mcp-port",
        help="Port for in-process MCP HTTP server (default: random free port).",
    ),
    dry_run: bool = typer.Option(
        False,
        "--dry-run",
        help="Print the generated assets and launch command without executing.",
    ),
    keep_temp: bool = typer.Option(
        False,
        "--keep-temp",
        help="Keep generated temporary MCP/skills files after exit.",
    ),
) -> None:
    """Launch an agent CLI with worai MCP and forward extra args after `--`."""
    if ctx.invoked_subcommand:
        return

    executable = _resolve_executable(agent_cli, agent_exec)
    _require_executable(executable)

    effective_transport = mcp_transport
    if agent_cli == AgentCli.gemini and mcp_transport == McpTransport.http:
        # Gemini MCP currently interoperates with worai via stdio transport.
        effective_transport = McpTransport.stdio

    resolved_config = str(resolve_config_path(config, env=os.environ)) if config else None
    running_http = None
    mcp_url: str | None = None
    if effective_transport == McpTransport.http:
        running_http = start_http_server(
            host=mcp_host,
            port=mcp_port,
            config_path=resolved_config,
            profile=profile,
        )
        mcp_url = running_http.url

    assets = _create_assets(config=resolved_config, profile=profile, mcp_url=mcp_url)
    passthrough = list(ctx.args)
    command, extra_env = _build_launch_command(
        agent_cli=agent_cli,
        agent_exec=executable,
        assets=assets,
        config=resolved_config,
        profile=profile,
        transport=effective_transport,
        mcp_url=mcp_url,
        passthrough=passthrough,
    )

    if dry_run:
        typer.echo(f"skills: {assets.skills_path}")
        typer.echo(f"context: {assets.wordlift_md_path}")
        typer.echo(f"mcp_config: {assets.mcp_config_path}")
        typer.echo(f"launch: {shlex.join(command)}")
        if running_http is not None:
            running_http.stop()
        if not keep_temp:
            typer.echo("dry-run: temporary files retained for inspection.")
        return

    if agent_cli == AgentCli.gemini:
        _print_wordlift_banner()

    env = os.environ.copy()
    env["WORAI_DISABLE_UPDATE_CHECK"] = "1"
    env.update(extra_env)
    if agent_cli == AgentCli.gemini:
        env.update(_gemini_env(assets.gemini_home))
    try:
        completed = subprocess.run(command, env=env, check=False)
    finally:
        if running_http is not None:
            running_http.stop()
        if not keep_temp:
            for path in (assets.skills_path, assets.wordlift_md_path, assets.mcp_config_path):
                try:
                    path.unlink()
                except OSError:
                    pass
            shutil.rmtree(assets.temp_dir, ignore_errors=True)
    raise typer.Exit(code=int(completed.returncode))


@mcp_app.command("serve", help="Run worai MCP server over stdio.")
def mcp_serve(
    config: str | None = typer.Option(
        None,
        "--config",
        help="Session default config path for MCP tools.",
    ),
    profile: str | None = typer.Option(
        None,
        "--profile",
        help="Session default profile for MCP tools.",
    ),
) -> None:
    resolved_config = str(resolve_config_path(config, env=os.environ)) if config else None
    raise typer.Exit(code=run_stdio_server(config_path=resolved_config, profile=profile))



"""Shared concurrency parsing and auto-scaling policy."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Iterable

from worai.errors import UsageError


@dataclass(frozen=True)
class ConcurrencySettings:
    mode: str
    workers: int
    min_workers: int
    max_workers: int

    @property
    def is_auto(self) -> bool:
        return self.mode == "auto"

    def with_workers(self, workers: int) -> ConcurrencySettings:
        if self.is_auto:
            bounded = max(self.min_workers, min(workers, self.max_workers))
            return ConcurrencySettings(
                mode=self.mode,
                workers=bounded,
                min_workers=self.min_workers,
                max_workers=self.max_workers,
            )
        return self

    def next_workers(self, buckets: Iterable[str]) -> int:
        if not self.is_auto:
            return self.workers

        bucket_set = set(buckets)
        if bucket_set & {"throttle", "server_error", "error"}:
            return max(self.min_workers, self.workers - 1)
        if bucket_set and bucket_set == {"ok"}:
            return min(self.max_workers, self.workers + 1)
        return self.workers


def parse_concurrency(
    value: str | int,
    *,
    auto_default_workers: int = 4,
    auto_min_workers: int = 2,
    auto_max_workers: int = 12,
) -> ConcurrencySettings:
    if isinstance(value, int):
        workers = value
    else:
        normalized = value.strip().lower()
        if normalized == "auto":
            workers = max(auto_min_workers, min(auto_default_workers, auto_max_workers))
            return ConcurrencySettings(
                mode="auto",
                workers=workers,
                min_workers=auto_min_workers,
                max_workers=auto_max_workers,
            )
        try:
            workers = int(normalized)
        except ValueError as exc:
            raise UsageError("Concurrency must be an integer or 'auto'.") from exc

    if workers <= 0:
        raise UsageError("Concurrency must be greater than 0.")

    return ConcurrencySettings(
        mode="fixed",
        workers=workers,
        min_workers=workers,
        max_workers=workers,
    )


def classify_status_bucket(status_code: int | None, *, had_error: bool = False) -> str:
    if had_error:
        return "error"
    if status_code is None:
        return "error"
    if status_code == 429:
        return "throttle"
    if 500 <= status_code < 600:
        return "server_error"
    if 200 <= status_code < 400:
        return "ok"
    return "client_error"



from __future__ import annotations

from pathlib import Path
from types import SimpleNamespace
from unittest.mock import AsyncMock, patch

from typer.testing import CliRunner

from worai.commands import graph as graph_cmd
from worai.errors import UsageError


def _patch_reset(monkeypatch, side_effect=None):
    mock = AsyncMock(return_value=None, side_effect=side_effect)
    monkeypatch.setattr(graph_cmd, "_reset_me", mock)
    return mock


def _patch_api_key(monkeypatch, api_key="wl_test"):
    monkeypatch.setattr(
        graph_cmd,
        "resolve_profile_api_key",
        lambda _ctx: (api_key, "default", Path("worai.toml")),
    )


# ---------------------------------------------------------------------------
# SDK unavailable
# ---------------------------------------------------------------------------


def test_graph_reset_raises_usage_error_when_sdk_unavailable(monkeypatch) -> None:
    monkeypatch.setattr(graph_cmd, "_reset_me", None)
    _patch_api_key(monkeypatch)

    result = CliRunner().invoke(graph_cmd.app, ["reset", "--yes"])

    assert result.exit_code == 1
    assert isinstance(result.exception, UsageError)
    assert "wordlift-sdk>=6.12.0" in str(result.exception)


# ---------------------------------------------------------------------------
# API key resolution
# ---------------------------------------------------------------------------


def test_graph_reset_raises_usage_error_when_api_key_missing(monkeypatch) -> None:
    _patch_reset(monkeypatch)
    _patch_api_key(monkeypatch, api_key=None)

    result = CliRunner().invoke(graph_cmd.app, ["reset", "--yes"])

    assert result.exit_code == 1
    assert isinstance(result.exception, UsageError)
    assert "WORDLIFT_API_KEY" in str(result.exception)


def test_graph_reset_raises_usage_error_on_profile_error(monkeypatch) -> None:
    _patch_reset(monkeypatch)
    monkeypatch.setattr(
        graph_cmd,
        "resolve_profile_api_key",
        lambda _ctx: (_ for _ in ()).throw(ValueError("no profile")),
    )

    result = CliRunner().invoke(graph_cmd.app, ["reset", "--yes"])

    assert result.exit_code == 1
    assert isinstance(result.exception, UsageError)
    assert "no profile" in str(result.exception)


# ---------------------------------------------------------------------------
# Confirmation prompt
# ---------------------------------------------------------------------------


def test_graph_reset_prompts_without_yes(monkeypatch) -> None:
    mock = _patch_reset(monkeypatch)
    _patch_api_key(monkeypatch)
    monkeypatch.setattr(graph_cmd.typer, "confirm", lambda *_a, **_k: True)

    result = CliRunner().invoke(graph_cmd.app, ["reset"])

    assert result.exit_code == 0
    mock.assert_awaited_once()


def test_graph_reset_skips_prompt_with_yes(monkeypatch) -> None:
    mock = _patch_reset(monkeypatch)
    _patch_api_key(monkeypatch)

    result = CliRunner().invoke(graph_cmd.app, ["reset", "--yes"])

    assert result.exit_code == 0
    mock.assert_awaited_once()
    assert "Graph reset complete." in result.output


def test_graph_reset_aborts_on_prompt_decline(monkeypatch) -> None:
    mock = _patch_reset(monkeypatch)
    _patch_api_key(monkeypatch)
    monkeypatch.setattr(
        graph_cmd.typer,
        "confirm",
        lambda *_a, **_k: (_ for _ in ()).throw(
            graph_cmd.typer.Abort()
        ),
    )

    result = CliRunner().invoke(graph_cmd.app, ["reset"])

    assert result.exit_code != 0
    mock.assert_not_awaited()


# ---------------------------------------------------------------------------
# Option forwarding
# ---------------------------------------------------------------------------


def test_graph_reset_forwards_keep_flags(monkeypatch) -> None:
    mock = _patch_reset(monkeypatch)
    _patch_api_key(monkeypatch)

    result = CliRunner().invoke(
        graph_cmd.app,
        ["reset", "--yes", "--keep-country", "--keep-language", "--keep-url"],
    )

    assert result.exit_code == 0
    _, kwargs = mock.call_args
    assert kwargs["keep_country"] is True
    assert kwargs["keep_language"] is True
    assert kwargs["keep_url"] is True


def test_graph_reset_default_flags_are_false(monkeypatch) -> None:
    mock = _patch_reset(monkeypatch)
    _patch_api_key(monkeypatch)

    result = CliRunner().invoke(graph_cmd.app, ["reset", "--yes"])

    assert result.exit_code == 0
    _, kwargs = mock.call_args
    assert kwargs["keep_country"] is False
    assert kwargs["keep_language"] is False
    assert kwargs["keep_url"] is False


# ---------------------------------------------------------------------------
# Error wrapping
# ---------------------------------------------------------------------------


def test_graph_reset_wraps_runtime_error(monkeypatch) -> None:
    _patch_reset(monkeypatch, side_effect=RuntimeError("API error"))
    _patch_api_key(monkeypatch)

    result = CliRunner().invoke(graph_cmd.app, ["reset", "--yes"])

    assert result.exit_code == 1
    assert isinstance(result.exception, UsageError)
    assert "Graph reset failed" in str(result.exception)
    assert "API error" in str(result.exception)


# ---------------------------------------------------------------------------
# Warning message
# ---------------------------------------------------------------------------


def test_graph_reset_prints_warning(monkeypatch) -> None:
    _patch_reset(monkeypatch)
    _patch_api_key(monkeypatch)

    result = CliRunner().invoke(graph_cmd.app, ["reset", "--yes"])

    assert "WARNING" in result.output
    assert "destructive" in result.output



"""Request/response models for structured data workflows."""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path


@dataclass
class CreateRequest:
    url: str
    target_type: str
    output_dir: Path
    base_name: str
    jsonld_path: Path | None
    yarrml_path: Path | None
    api_key: str | None
    base_url: str | None
    debug: bool
    headed: bool
    timeout_ms: int
    max_retries: int
    quality_check: bool
    max_xhtml_chars: int
    max_text_node_chars: int
    max_nesting_depth: int
    verbose: bool
    validate: bool
    wait_until: str


@dataclass
class GenerateRequest:
    input_value: str
    yarrrml_path: Path
    regex: str
    output_dir: Path
    output_format: str
    concurrency: str
    api_key: str | None
    base_url: str | None
    headed: bool
    timeout_ms: int
    wait_until: str
    max_xhtml_chars: int
    max_text_node_chars: int
    max_pages: int | None
    verbose: bool



"""Unified output renderer for worai commands.

Usage::

    from worai.core.output import OutputFormat, write_output

    write_output(rows, OutputFormat.csv, dest=Path("out.csv"))
    write_output(rows, OutputFormat.table)           # → stdout rich table
    write_output(rows, OutputFormat.json)            # → stdout JSON array
    write_output(rows, OutputFormat.parquet, dest=Path("out.parquet"))
    write_output(
        rows, OutputFormat.gsheets,
        credentials_file="svc.json",
        spreadsheet_id="<id>",
        sheet_name="Sheet1",
    )
"""

from __future__ import annotations

import csv
import json
import sys
from enum import Enum
from pathlib import Path
from typing import Any


class OutputFormat(str, Enum):
    table = "table"
    json = "json"
    csv = "csv"
    tsv = "tsv"
    parquet = "parquet"
    gsheets = "gsheets"


def write_output(
    rows: list[dict],
    fmt: OutputFormat,
    dest: str | Path | None = None,
    **fmt_options: Any,
) -> None:
    """Write *rows* in the requested *fmt* to *dest* (stdout when None).

    Args:
        rows:        Uniformly-keyed dicts; column order is inferred from the
                     first row's key order.
        fmt:         One of the :class:`OutputFormat` values.
        dest:        ``None`` → stdout (not valid for ``parquet``);
                     path string or :class:`~pathlib.Path` → write to file.
        **fmt_options:
            ``no_color``        (table)  – strip ANSI codes; default False.
            ``credentials_file`` (gsheets) – service-account JSON path.
            ``spreadsheet_id``   (gsheets) – Google Sheets document ID.
            ``sheet_name``       (gsheets) – worksheet tab name.
    """
    if fmt == OutputFormat.table:
        _write_table(rows, dest, no_color=fmt_options.get("no_color", False))
    elif fmt == OutputFormat.json:
        _write_json(rows, dest)
    elif fmt == OutputFormat.csv:
        _write_separated(rows, dest, delimiter=",")
    elif fmt == OutputFormat.tsv:
        _write_separated(rows, dest, delimiter="\t")
    elif fmt == OutputFormat.parquet:
        _write_parquet(rows, dest)
    elif fmt == OutputFormat.gsheets:
        _write_gsheets(rows, **fmt_options)
    else:
        raise ValueError(f"Unsupported output format: {fmt!r}")


# ---------------------------------------------------------------------------
# Format handlers
# ---------------------------------------------------------------------------


def _write_table(rows: list[dict], dest: str | Path | None, *, no_color: bool = False) -> None:
    from rich.console import Console
    from rich.table import Table

    if not rows:
        return

    if dest is not None:
        f = open(dest, "w", encoding="utf-8")
        console = Console(file=f, no_color=True)
    else:
        console = Console(file=sys.stdout, no_color=no_color)

    columns = list(rows[0].keys())
    table = Table(*columns)
    for row in rows:
        table.add_row(*[str(row.get(col, "")) for col in columns])

    try:
        console.print(table)
    finally:
        if dest is not None:
            f.close()


def _write_json(rows: list[dict], dest: str | Path | None) -> None:
    text = json.dumps(rows, indent=2)
    if dest is None:
        sys.stdout.write(text + "\n")
    else:
        Path(dest).write_text(text, encoding="utf-8")


def _write_separated(rows: list[dict], dest: str | Path | None, *, delimiter: str) -> None:
    fieldnames: list[str] = list(rows[0].keys()) if rows else []

    if dest is None:
        writer = csv.DictWriter(sys.stdout, fieldnames=fieldnames, delimiter=delimiter)
        writer.writeheader()
        writer.writerows(rows)
    else:
        with open(dest, "w", newline="", encoding="utf-8") as f:
            writer = csv.DictWriter(f, fieldnames=fieldnames, delimiter=delimiter)
            writer.writeheader()
            writer.writerows(rows)


def _write_parquet(rows: list[dict], dest: str | Path | None) -> None:
    if dest is None:
        raise ValueError(
            "parquet format requires a destination file path "
            "(pass --output / dest= to write_output)."
        )
    try:
        import pyarrow as pa
        import pyarrow.parquet as pq
    except ImportError:
        raise ImportError(
            "pyarrow is required for parquet output. "
            "Install it with: pip install 'worai[parquet]'"
        ) from None

    table = pa.Table.from_pylist(rows)
    pq.write_table(table, str(dest))


def _write_gsheets(rows: list[dict], **fmt_options: Any) -> None:
    credentials_file = fmt_options.get("credentials_file")
    spreadsheet_id = fmt_options.get("spreadsheet_id")
    sheet_name = fmt_options.get("sheet_name")

    missing = [k for k, v in [
        ("credentials_file", credentials_file),
        ("spreadsheet_id", spreadsheet_id),
        ("sheet_name", sheet_name),
    ] if not v]
    if missing:
        raise ValueError(
            f"gsheets format requires fmt_options: {', '.join(missing)}"
        )

    try:
        import gspread
    except ImportError:
        raise ImportError(
            "gspread is required for gsheets output. "
            "Install it with: pip install 'worai[gsheets]'"
        ) from None

    gc = gspread.service_account(filename=credentials_file)
    spreadsheet = gc.open_by_key(spreadsheet_id)

    try:
        worksheet = spreadsheet.worksheet(sheet_name)
    except gspread.exceptions.WorksheetNotFound:
        cols = len(rows[0]) if rows else 1
        worksheet = spreadsheet.add_worksheet(
            title=sheet_name, rows=len(rows) + 1, cols=cols
        )

    worksheet.clear()

    if not rows:
        return

    headers = list(rows[0].keys())
    data = [headers] + [[row.get(col, "") for col in headers] for row in rows]
    worksheet.update(data)



"""CLI wrapper for entity-matrix."""

from __future__ import annotations

from pathlib import Path
from typing import Optional

import typer

from worai.core.config_resolver import (
    resolve_gsheets_credentials,
    resolve_gsheets_sheet_name,
    resolve_gsheets_spreadsheet_id,
)
from worai.core.profile_loader import load_raw_profile
from worai.core.entity_matrix import EntityMatrixOptions, build_entity_matrix
from worai.core.output import OutputFormat, write_output
from worai.errors import UsageError


def _to_presence(rows: list[dict]) -> list[dict]:
    """Replace integer counts with ✅ (present) or '' (absent/zero)."""
    return [
        {k: ("✅" if v else "") if k != "url" else v for k, v in row.items()}
        for row in rows
    ]


def run(
    ctx: typer.Context,
    filename: str = typer.Argument(..., help="Path to Turtle (.ttl) graph file."),
    exclude_type: Optional[list[str]] = typer.Option(
        None,
        "--exclude-type",
        "-x",
        help="Entity type short name to exclude (e.g. WebPage). Repeat to exclude multiple.",
    ),
    cluster: bool = typer.Option(
        False,
        "--cluster/--no-cluster",
        help=(
            "Aggregate URL rows that share a common parent path and identical "
            "entity-type set into a single wildcard row (e.g. /blog/*)."
        ),
    ),
    depth: int = typer.Option(
        1,
        "--depth",
        help="Subgraph relationship hops for referenced entity traversal.",
        show_default=True,
    ),
    count: bool = typer.Option(
        False,
        "--count/--no-count",
        help="Show integer counts instead of ✅ presence markers.",
    ),
    fmt: OutputFormat = typer.Option(OutputFormat.table, "--format", "-f", help="Output format."),
    output: Optional[Path] = typer.Option(None, "--output", "-o", help="Output file path."),
    gsheets_credentials: Optional[str] = typer.Option(
        None,
        "--gsheets-credentials",
        help="Service-account JSON path (required for gsheets format).",
        show_default=False,
    ),
    gsheets_id: Optional[str] = typer.Option(
        None,
        "--gsheets-id",
        help="Google Sheets spreadsheet ID (required for gsheets format).",
        show_default=False,
    ),
    gsheets_sheet: Optional[str] = typer.Option(
        None,
        "--gsheets-sheet",
        help="Worksheet tab name (required for gsheets format).",
        show_default=False,
    ),
) -> None:
    try:
        options = EntityMatrixOptions(
            filename=filename,
            exclude_types=list(exclude_type) if exclude_type else [],
            cluster=cluster,
            subgraph_depth=depth,
        )
        rows = build_entity_matrix(options)
        if not count:
            rows = _to_presence(rows)
    except FileNotFoundError:
        raise UsageError(f"File not found: '{filename}'")
    except Exception as exc:
        raise UsageError(f"Failed to build entity matrix: {exc}") from exc

    fmt_options: dict = {}
    if fmt == OutputFormat.gsheets:
        raw = load_raw_profile(ctx)
        creds = resolve_gsheets_credentials(gsheets_credentials, raw)
        shid = resolve_gsheets_spreadsheet_id(gsheets_id, raw)
        sheet = resolve_gsheets_sheet_name(gsheets_sheet, raw)
        if creds:
            fmt_options["credentials_file"] = creds
        if shid:
            fmt_options["spreadsheet_id"] = shid
        if sheet:
            fmt_options["sheet_name"] = sheet

    write_output(rows, fmt, dest=output, **fmt_options)



from __future__ import annotations

from worai.core.http import DEFAULT_USER_AGENT, build_default_user_agent


def test_default_user_agent_contains_wlbot_worai() -> None:
    ua = build_default_user_agent()
    assert "WLBot/worai-" in ua
    assert "contact: support@wordlift.io" in ua
    assert "Chrome/" in ua


def test_default_user_agent_constant_matches_builder() -> None:
    assert DEFAULT_USER_AGENT == build_default_user_agent()



from __future__ import annotations

from types import SimpleNamespace

import pytest

from worai.core.url_sources import sheets
from worai.errors import UsageError


def test_extract_and_is_spreadsheet_source() -> None:
    assert sheets.extract_spreadsheet_id("https://docs.google.com/spreadsheets/d/abc123-XYZ_/edit") == "abc123-XYZ_"
    assert sheets.extract_spreadsheet_id("a" * 25) == "a" * 25
    assert sheets.extract_spreadsheet_id("not-a-sheet") is None
    assert sheets.is_spreadsheet_source("a" * 25) is True
    assert sheets.is_spreadsheet_source("x") is False


def test_load_urls_from_sheet_paths(monkeypatch: pytest.MonkeyPatch) -> None:
    class _Session:
        def __init__(self, _creds):
            pass

        def get(self, url: str):
            if "missing-url-col" in url:
                return SimpleNamespace(status_code=200, json=lambda: {"values": [["name"], ["x"]]})
            if "empty" in url:
                return SimpleNamespace(status_code=200, json=lambda: {"values": []})
            if "bad-status" in url:
                return SimpleNamespace(status_code=500, text="boom", json=lambda: {})
            return SimpleNamespace(
                status_code=200,
                json=lambda: {
                    "values": [["url", "title"], [" https://a.com ", "A"], ["", "B"], ["https://b.com", "B"]]
                },
            )

    monkeypatch.setattr(sheets, "load_google_credentials", lambda **_k: object())
    monkeypatch.setattr(sheets, "AuthorizedSession", _Session)

    with pytest.raises(UsageError, match="Invalid Google Sheets source"):
        sheets.load_urls_from_sheet("invalid", "Sheet1", client_secrets=None, token="t.json", port=8080)

    with pytest.raises(RuntimeError, match="Failed to read source spreadsheet values"):
        sheets.load_urls_from_sheet("bad-status" + "x" * 20, "Sheet1", client_secrets=None, token="t.json", port=8080)

    with pytest.raises(UsageError, match="missing required 'url' column"):
        sheets.load_urls_from_sheet("missing-url-col" + "x" * 20, "Sheet1", client_secrets=None, token="t.json", port=8080)

    assert sheets.load_urls_from_sheet("empty" + "x" * 20, "Sheet1", client_secrets=None, token="t.json", port=8080) == []

    urls = sheets.load_urls_from_sheet("ok" + "x" * 20, "Sheet1", client_secrets=None, token="t.json", port=8080)
    assert urls == ["https://a.com", "https://b.com"]



from __future__ import annotations

import json
from types import SimpleNamespace

from typer.testing import CliRunner

from worai.commands import graph as graph_cmd
from worai.errors import UsageError


class _Result:
    def __init__(self, conforms: bool = True, warning_count: int = 0) -> None:
        self.conforms = conforms
        self.warning_count = warning_count


def test_graph_validate_uses_shape_resolver_and_validate_file(monkeypatch) -> None:
    seen = {}

    def _resolve(**kwargs):
        seen["resolve"] = kwargs
        return ["resolved-shape.ttl"]

    def _validate_file(input_file: str, shape_specs=None):
        seen.setdefault("calls", []).append((input_file, shape_specs))
        return _Result(conforms=True, warning_count=0)

    monkeypatch.setattr(graph_cmd, "_sdk_resolve_shape_specs", _resolve)
    monkeypatch.setattr(graph_cmd, "_sdk_validate_file", _validate_file)
    monkeypatch.setattr(graph_cmd, "_sdk_extract_validation_issues", lambda _r: [])
    monkeypatch.setattr(graph_cmd, "_sdk_filter_validation_issues", lambda issues, level: issues)

    result = CliRunner().invoke(
        graph_cmd.app,
        [
            "validate",
            "graph.ttl",
            "--builtin-shape",
            "google-required",
            "--exclude-builtin-shape",
            "schemaorg-grammar",
            "--shape",
            "./custom.ttl",
        ],
    )

    assert result.exit_code == 0
    assert seen["resolve"]["builtin_shapes"] == ["google-required"]
    assert seen["resolve"]["exclude_builtin_shapes"] == ["schemaorg-grammar"]
    assert seen["resolve"]["extra_shapes"] == ["./custom.ttl"]
    assert seen["calls"] == [("graph.ttl", ["resolved-shape.ttl"])]


def test_graph_validate_fails_at_warning_level(monkeypatch) -> None:
    monkeypatch.setattr(graph_cmd, "_sdk_resolve_shape_specs", lambda **_k: [])
    monkeypatch.setattr(graph_cmd, "_sdk_validate_file", lambda *_a, **_k: _Result(conforms=True, warning_count=1))
    issue = SimpleNamespace(
        level="warning",
        severity="Warning",
        focus_node="urn:focus",
        result_path="schema:name",
        rule_id="r1",
        rule_set="google",
        message="Missing recommended property",
    )
    monkeypatch.setattr(graph_cmd, "_sdk_extract_validation_issues", lambda _r: [issue])
    monkeypatch.setattr(
        graph_cmd,
        "_sdk_filter_validation_issues",
        lambda issues, level: issues if level == "warning" else [],
    )

    result = CliRunner().invoke(graph_cmd.app, ["validate", "graph.ttl", "--level", "warning"])

    assert result.exit_code == 1
    assert isinstance(result.exception, UsageError)
    assert "FAIL" in result.output


def test_graph_validate_error_level_ignores_warning_issues(monkeypatch) -> None:
    monkeypatch.setattr(graph_cmd, "_sdk_resolve_shape_specs", lambda **_k: [])
    monkeypatch.setattr(graph_cmd, "_sdk_validate_file", lambda *_a, **_k: _Result(conforms=True, warning_count=1))
    issue = SimpleNamespace(
        level="warning",
        severity="Warning",
        focus_node=None,
        result_path=None,
        rule_id="r1",
        rule_set="google",
        message="Missing recommended property",
    )
    monkeypatch.setattr(graph_cmd, "_sdk_extract_validation_issues", lambda _r: [issue])
    monkeypatch.setattr(
        graph_cmd,
        "_sdk_filter_validation_issues",
        lambda issues, level: issues if level == "warning" else [],
    )

    result = CliRunner().invoke(graph_cmd.app, ["validate", "graph.ttl", "--level", "error"])

    assert result.exit_code == 0
    assert "graph.ttl: OK" in result.output


def test_graph_validate_json_output_for_multiple_inputs(monkeypatch) -> None:
    monkeypatch.setattr(graph_cmd, "_sdk_resolve_shape_specs", lambda **_k: [])
    monkeypatch.setattr(graph_cmd, "_sdk_validate_file", lambda *_a, **_k: _Result(conforms=True, warning_count=0))
    monkeypatch.setattr(graph_cmd, "_sdk_extract_validation_issues", lambda _r: [])
    monkeypatch.setattr(graph_cmd, "_sdk_filter_validation_issues", lambda issues, level: issues)

    result = CliRunner().invoke(graph_cmd.app, ["validate", "a.ttl", "b.ttl", "--format", "json"])

    assert result.exit_code == 0
    payload = json.loads(result.output)
    assert isinstance(payload, list)
    assert [item["input"] for item in payload] == ["a.ttl", "b.ttl"]
    assert all(item["ok"] for item in payload)


def test_graph_validate_builtin_shapes_require_sdk_610(monkeypatch) -> None:
    monkeypatch.setattr(graph_cmd, "_sdk_resolve_shape_specs", None)

    result = CliRunner().invoke(graph_cmd.app, ["validate", "graph.ttl", "--builtin-shape", "google-required"])

    assert result.exit_code == 1
    assert isinstance(result.exception, UsageError)
    assert "wordlift-sdk>=6.1.0" in str(result.exception)



from __future__ import annotations

import datetime as dt
import types

import pandas as pd
import pytest

import worai.seoreport.core as core


def test_get_period_ranges_math() -> None:
    last = dt.date(2026, 2, 24)
    ranges = core.get_period_ranges(last, 7)
    assert ranges.current.end == last
    assert ranges.current.start == dt.date(2026, 2, 18)
    assert ranges.pop.end == dt.date(2026, 2, 17)


def test_fetch_gsc_data_paginates(monkeypatch: pytest.MonkeyPatch) -> None:
    calls = {"n": 0}

    def _stub(_session, _site, payload):
        calls["n"] += 1
        if payload["startRow"] == 0:
            return {"rows": [{"keys": ["q1"], "clicks": 1, "impressions": 2, "ctr": 0.5, "position": 3}]}
        return {"rows": []}

    monkeypatch.setattr(core, "gsc_post", _stub)
    df = core.fetch_gsc_data(object(), "s", dt.date(2026, 1, 1), dt.date(2026, 1, 2), ["query"], row_limit=1)
    assert calls["n"] == 2
    assert list(df["query"]) == ["q1"]


def test_run_with_retries(monkeypatch: pytest.MonkeyPatch) -> None:
    count = {"n": 0}
    monkeypatch.setattr(core.time, "sleep", lambda _x: None)
    monkeypatch.setattr(core.random, "random", lambda: 0.0)

    def _flaky():
        count["n"] += 1
        if count["n"] < 2:
            raise RuntimeError("x")
        return 7

    assert core._run_with_retries(_flaky, retries=2) == 7


def test_aggregate_and_diffs() -> None:
    df = pd.DataFrame(
        [
            {"query": "q", "page": "/a", "clicks": 2, "impressions": 4, "ctr": 0.5, "position": 4},
            {"query": "q", "page": "/b", "clicks": 1, "impressions": 1, "ctr": 1.0, "position": 2},
        ]
    )
    out = core.aggregate_data(df, None)
    assert out.loc["q", "clicks"] == 3
    assert out.loc["q", "impressions"] == 5

    diffs = core.calculate_diffs(pd.Series({"clicks": 10, "impressions": 0, "ctr": 0.0, "position": 1.0}), pd.Series())
    assert diffs["abs"]["clicks"] == 10


def test_fetch_inspection_and_health(monkeypatch: pytest.MonkeyPatch) -> None:
    class _Resp:
        status_code = 200

        @staticmethod
        def json():
            return {"inspectionResult": {"indexStatusResult": {"verdict": "FAIL", "coverageState": "Soft 404"}}}

        text = "ok"

    class _Session:
        @staticmethod
        def post(_url, json=None):
            return _Resp()

    monkeypatch.setattr(core.time, "sleep", lambda _x: None)
    summary = core.fetch_inspection_data(_Session(), "site", ["https://example.com/a"])
    assert summary.error == 1
    assert "Soft 404" in summary.issues

    class _HeadResp:
        status_code = 405

    class _GetResp:
        status_code = 200

        @staticmethod
        def close():
            return None

    monkeypatch.setattr(core.requests, "head", lambda *_a, **_k: _HeadResp())
    monkeypatch.setattr(core.requests, "get", lambda *_a, **_k: _GetResp())
    seq = iter([1.0, 1.2, 2.0, 2.1])
    monkeypatch.setattr(core.time, "time", lambda: next(seq))
    health = core.run_live_health_check(["https://example.com/a", "https://example.com/b"])
    assert health.total_checked == 2
    assert health.error_pages == 0


def test_analyze_period_with_stubbed_fetch(monkeypatch: pytest.MonkeyPatch) -> None:
    date = dt.date(2026, 2, 24)
    ranges = core.get_period_ranges(date, 7)

    curr = pd.DataFrame([{"query": "q", "clicks": 10, "impressions": 20, "ctr": 0.5, "position": 4.0}])
    prev = pd.DataFrame([{"query": "q", "clicks": 5, "impressions": 10, "ctr": 0.5, "position": 5.0}])
    yoy = pd.DataFrame([{"query": "q", "clicks": 4, "impressions": 9, "ctr": 0.4, "position": 6.0}])
    trend = pd.DataFrame([{"date": "2026-02-24", "clicks": 1, "impressions": 2, "ctr": 0.5, "position": 3.0}])

    monkeypatch.setattr(core, "_run_tasks", lambda _tasks, max_workers: {"curr": curr, "pop": prev, "yoy": yoy, "trend": trend})
    monkeypatch.setattr(core, "AuthorizedSession", lambda _creds: object())
    analysis = core.analyze_period("Last 7 days", ranges, creds=object(), site="s", url_regex=None, max_workers=2)
    assert analysis.name == "Last 7 days"
    assert analysis.overall["current"]["clicks"] == 10
    assert analysis.daily_trend[0].date == "2026-02-24"


def test_generate_report_data_with_ga(monkeypatch: pytest.MonkeyPatch) -> None:
    date = dt.date(2026, 2, 24)

    ranges = core.PeriodRanges(
        current=core.DateRange(date, date),
        pop=core.DateRange(date, date),
        yoy=core.DateRange(date, date),
    )

    period = core.PeriodAnalysis(name="Last 7 days", dates=ranges, overall={}, winning=[], losing=[], opportunities=[])
    monkeypatch.setattr(core, "load_credentials", lambda *_a, **_k: object())
    monkeypatch.setattr(core, "AuthorizedSession", lambda _c: object())
    monkeypatch.setattr(core, "find_last_available_date", lambda _s, _site: date)
    monkeypatch.setattr(core, "_run_tasks", lambda _tasks, max_workers: {"7": period, "30": period, "90": period})
    monkeypatch.setattr(core, "build_analytics", lambda *_a, **_k: {"periods": [{"name": "Last 7 days"}]})
    monkeypatch.setattr(core, "load_ga_credentials", lambda *_a, **_k: object())
    monkeypatch.setattr(core, "_resolve_ga_auth", lambda _opts: ("ga.json", "tok.json"))

    data = core.generate_report_data(
        core.ReportOptions(site="s", ga_id="123", client_secrets="c.json", token="t.json", inspect_limit=0)
    )
    assert data["analytics"] is not None



from __future__ import annotations

from pathlib import Path
import sys
import types

import pytest
from typer.testing import CliRunner

from worai import cli as root_cli
from worai.commands import web_pages as web_pages_cmd
from worai.core.web_pages import WebPagesClassifyTypesOptions, run_classify_types
from worai.errors import UsageError


def _install_fake_sdk_ingestion(monkeypatch: pytest.MonkeyPatch, callback) -> None:
    pkg = types.ModuleType("wordlift_sdk")
    mod = types.ModuleType("wordlift_sdk.ingestion")
    mod.create_type_classification_csv_from_ingestion = callback
    pkg.ingestion = mod
    monkeypatch.setitem(sys.modules, "wordlift_sdk", pkg)
    monkeypatch.setitem(sys.modules, "wordlift_sdk.ingestion", mod)


def test_run_classify_types_urls_source_bundle(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    seen: dict[str, object] = {}

    def _fake_create(**kwargs):
        seen["kwargs"] = kwargs
        return [1]

    _install_fake_sdk_ingestion(monkeypatch, _fake_create)

    source = tmp_path / "urls.txt"
    source.write_text("https://example.com/a\nhttps://example.com/b\n", encoding="utf-8")

    result = run_classify_types(
        WebPagesClassifyTypesOptions(
            source=str(source),
            output_csv=str(tmp_path / "out.csv"),
            ingest_source="urls",
            ingest_loader="playwright",
            url_regex=r"/a$",
            agent_cli="codex",
            agent_timeout_sec=90.0,
            max_markdown_chars=9000,
        )
    )

    assert result == [1]
    kwargs = seen["kwargs"]
    assert kwargs["output_csv"] == str(tmp_path / "out.csv")
    assert kwargs["agent_cli"] == "codex"
    assert kwargs["agent_timeout_sec"] == 90.0
    assert kwargs["max_markdown_chars"] == 9000
    bundle = kwargs["source_bundle"]
    assert bundle["INGEST_SOURCE"] == "urls"
    assert bundle["INGEST_LOADER"] == "playwright"
    assert bundle["URL_REGEX"] == r"/a$"
    assert bundle["URLS"] == ["https://example.com/a", "https://example.com/b"]


def test_run_classify_types_auto_infers_sitemap(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    seen: dict[str, object] = {}

    def _fake_create(**kwargs):
        seen["kwargs"] = kwargs
        return []

    _install_fake_sdk_ingestion(monkeypatch, _fake_create)

    run_classify_types(
        WebPagesClassifyTypesOptions(
            source="https://example.com/sitemap.xml",
            output_csv=str(tmp_path / "out.csv"),
            ingest_source=None,
            ingest_loader=None,
        )
    )
    bundle = seen["kwargs"]["source_bundle"]
    assert bundle["INGEST_SOURCE"] == "sitemap"
    assert bundle["INGEST_LOADER"] == "web_scrape_api"
    assert "INGEST_TIMEOUT_MS" not in bundle
    assert "PLAYWRIGHT_WAIT_UNTIL" not in bundle
    assert bundle["SITEMAP_URL"] == "https://example.com/sitemap.xml"


def test_run_classify_types_sheets_requires_sheet_and_service_account(tmp_path: Path) -> None:
    with pytest.raises(UsageError, match="--sheet-name is required"):
        run_classify_types(
            WebPagesClassifyTypesOptions(
                source="https://docs.google.com/spreadsheets/d/12345678901234567890/edit",
                output_csv=str(tmp_path / "out.csv"),
                ingest_source="sheets",
            )
        )

    with pytest.raises(UsageError, match="SHEETS_SERVICE_ACCOUNT is required"):
        run_classify_types(
            WebPagesClassifyTypesOptions(
                source="https://docs.google.com/spreadsheets/d/12345678901234567890/edit",
                output_csv=str(tmp_path / "out.csv"),
                ingest_source="sheets",
                sheet_name="URLs",
            )
        )


def test_web_pages_command_wires_options(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    seen: dict[str, object] = {}

    def _fake_run(options, on_progress=None):
        assert callable(on_progress)
        seen["options"] = options
        return [1, 2]

    monkeypatch.setattr(web_pages_cmd, "run_classify_types", _fake_run)
    monkeypatch.setattr(
        web_pages_cmd,
        "load_profile_settings",
        lambda _ctx: (
            {
                "ingest.timeout_ms": "9000",
                "ingest.retry_attempts": "2",
                "ingest.retry_backoff_ms": "100",
                "ingest.playwright_wait_until": "domcontentloaded",
            },
            "default",
        ),
    )

    result = CliRunner().invoke(
        root_cli.app,
        [
            "web-pages",
            "classify-types",
            "https://example.com/sitemap.xml",
            "-y",
            "--ingest-source",
            "sitemap",
            "--ingest-loader",
            "playwright",
            "--url-regex",
            "car-insurance",
            "--output",
            str(tmp_path / "types.csv"),
        ],
    )

    assert result.exit_code == 0
    assert "Wrote 2 rows" in result.output
    options = seen["options"]
    assert options.source == "https://example.com/sitemap.xml"
    assert options.ingest_source == "sitemap"
    assert options.ingest_loader == "playwright"
    assert options.url_regex == "car-insurance"
    assert options.ingest_timeout_ms == 9000
    assert options.ingest_retry_attempts == 2
    assert options.ingest_retry_backoff_ms == 100
    assert options.playwright_wait_until == "domcontentloaded"


def test_root_cli_includes_web_pages_command() -> None:
    result = CliRunner().invoke(root_cli.app, ["--help"])
    assert result.exit_code == 0
    assert "web-pages" in result.output
    assert "Run web pages ingestion workflows." in result.output


def test_web_pages_classify_types_confirmation_abort(monkeypatch: pytest.MonkeyPatch) -> None:
    called = {"value": False}

    def _fake_run(_options, on_progress=None):
        called["value"] = True
        return []

    monkeypatch.setattr(web_pages_cmd, "run_classify_types", _fake_run)
    monkeypatch.setattr(web_pages_cmd, "load_profile_settings", lambda _ctx: ({}, "default"))

    result = CliRunner().invoke(
        root_cli.app,
        ["web-pages", "classify-types", "https://example.com/sitemap.xml"],
        input="n\n",
    )

    assert result.exit_code == 0
    assert "consume agent credits" in result.output
    assert "Aborted." in result.output
    assert called["value"] is False


def test_web_pages_classify_types_confirmation_default_yes(monkeypatch: pytest.MonkeyPatch) -> None:
    called = {"value": False}

    def _fake_run(_options, on_progress=None):
        assert callable(on_progress)
        called["value"] = True
        return [1]

    monkeypatch.setattr(web_pages_cmd, "run_classify_types", _fake_run)
    monkeypatch.setattr(web_pages_cmd, "load_profile_settings", lambda _ctx: ({}, "default"))

    result = CliRunner().invoke(
        root_cli.app,
        ["web-pages", "classify-types", "https://example.com/sitemap.xml"],
        input="\n",
    )

    assert result.exit_code == 0
    assert "consume agent credits" in result.output
    assert "Wrote 1 rows" in result.output
    assert called["value"] is True


def test_classify_progress_callback_lifecycle_and_zero_case() -> None:
    bars: list[object] = []

    class _Bar:
        def __init__(self, *, total, desc, unit, leave):
            self.total = total
            self.desc = desc
            self.unit = unit
            self.leave = leave
            self.updates: list[int] = []
            self.closed = False

        def update(self, delta: int):
            self.updates.append(delta)

        def refresh(self):
            return None

        def close(self):
            self.closed = True

    def _factory(**kwargs):
        bar = _Bar(**kwargs)
        bars.append(bar)
        return bar

    callback, close = web_pages_cmd._build_classify_progress_callback(
        enabled=True,
        progress_factory=_factory,
    )
    callback({"event": "type_classification.progress.started", "meta": {"total": 2}})
    callback({"event": "type_classification.progress.updated", "meta": {"completed": 1, "total": 2}})
    callback({"event": "type_classification.progress.updated", "meta": {"completed": 1, "total": 2}})
    callback({"event": "type_classification.progress.updated", "meta": {"completed": 2, "total": 2}})
    callback({"event": "type_classification.progress.completed", "meta": {"completed": 2, "total": 2}})
    close()

    assert len(bars) == 1
    assert bars[0].updates == [1, 1]
    assert bars[0].closed is True

    zero_bars: list[object] = []

    def _zero_factory(**kwargs):
        bar = _Bar(**kwargs)
        zero_bars.append(bar)
        return bar

    callback, close = web_pages_cmd._build_classify_progress_callback(
        enabled=True,
        progress_factory=_zero_factory,
    )
    callback({"event": "type_classification.progress.started", "meta": {"total": 0}})
    callback({"event": "type_classification.progress.updated", "meta": {"completed": 0, "total": 0}})
    callback({"event": "type_classification.progress.completed", "meta": {"completed": 0, "total": 0}})
    close()

    assert len(zero_bars) == 1
    assert zero_bars[0].updates == []
    assert zero_bars[0].closed is True


def test_web_pages_classify_types_error_update_logged_before_raise(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    def _fake_run(_options, on_progress=None):
        assert callable(on_progress)
        on_progress({"event": "type_classification.progress.started", "meta": {"total": 1}})
        on_progress(
            {
                "event": "type_classification.progress.updated",
                "meta": {
                    "status": "error",
                    "url": "https://example.com/page",
                    "error_type": "ExtractionError",
                    "error_message": "Could not parse content",
                    "completed": 1,
                    "total": 1,
                },
            }
        )
        raise UsageError("classification failed")

    monkeypatch.setattr(web_pages_cmd, "run_classify_types", _fake_run)
    monkeypatch.setattr(web_pages_cmd, "load_profile_settings", lambda _ctx: ({}, "default"))

    result = CliRunner().invoke(
        root_cli.app,
        ["web-pages", "classify-types", "https://example.com/sitemap.xml", "-y"],
    )

    assert result.exit_code != 0
    assert "Type classification failed url=https://example.com/page error_type=ExtractionError error_message=Could not parse content" in result.output



"""Canonical URL deduplication using GSC impressions from wordlift-sdk."""

from __future__ import annotations

import json
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

from google.auth.credentials import Credentials
from google.auth.transport.requests import Request
from google.oauth2 import service_account

from worai.core.gsc import load_credentials
from worai.errors import UsageError

if TYPE_CHECKING:  # pragma: no cover
    import pandas as pd

WEBMASTERS_READONLY_SCOPE = "https://www.googleapis.com/auth/webmasters.readonly"

_AUTH_HELP_MESSAGE = """Authentication is missing.

Provide one of:
1) --service-account <path-or-json>   (recommended for CI/non-interactive)
2) --client-secrets <client.json>     (to bootstrap OAuth token)
3) --token <oauth_token.json>         (if already created and valid)

How to create credentials:
- Search Console API auth requirements:
  https://developers.google.com/webmaster-tools/v1/how-tos/authorizing
- Create OAuth client (Desktop app):
  https://developers.google.com/workspace/guides/create-credentials
- Create service account:
  https://cloud.google.com/iam/docs/service-accounts-create
- Add account to Search Console property permissions:
  https://support.google.com/webmasters/answer/7687615
"""


@dataclass
class CanonicalsDedupeOptions:
    input_csv: str
    output_csv: str
    site_url: str
    interval: str = "28d"
    url_regex: str | None = None
    concurrency: str = "auto"
    request_timeout_sec: float = 30.0
    service_account: str | None = None
    client_secrets: str | None = None
    token: str = "oauth_token.json"
    oauth_port: int = 8080


def _load_oauth_credentials(token_path: str, client_secrets: str | None, port: int) -> Credentials:
    token = Path(token_path)
    if token.exists():
        from wordlift_sdk.google_search_console import load_authorized_user_credentials

        try:
            creds = load_authorized_user_credentials(token)
            if creds.expired and creds.refresh_token:
                creds.refresh(Request())
                token.write_text(creds.to_json(), encoding="utf-8")
            if creds.valid:
                return creds
        except Exception:
            pass

    if not client_secrets:
        raise UsageError(_AUTH_HELP_MESSAGE)
    return load_credentials(client_secrets, token_path, port)


def _load_service_account_credentials(service_account_value: str) -> Credentials:
    as_path = Path(service_account_value)
    if as_path.exists():
        return service_account.Credentials.from_service_account_file(
            str(as_path),
            scopes=[WEBMASTERS_READONLY_SCOPE],
        )
    try:
        info = json.loads(service_account_value)
    except json.JSONDecodeError as exc:
        raise UsageError(
            "--service-account must be a valid file path or a JSON body."
        ) from exc
    if not isinstance(info, dict):
        raise UsageError("--service-account JSON must be an object.")
    return service_account.Credentials.from_service_account_info(
        info,
        scopes=[WEBMASTERS_READONLY_SCOPE],
    )


def run_dedupe(options: CanonicalsDedupeOptions) -> "pd.DataFrame":
    from wordlift_sdk.google_search_console import create_canonical_csv_from_gsc_impressions

    if options.service_account:
        credentials = _load_service_account_credentials(options.service_account)
        return create_canonical_csv_from_gsc_impressions(
            input_csv=options.input_csv,
            output_csv=options.output_csv,
            site_url=options.site_url,
            credentials=credentials,
            interval=options.interval,
            url_regex=options.url_regex,
            concurrency=options.concurrency,
            request_timeout_sec=options.request_timeout_sec,
        )

    credentials = _load_oauth_credentials(
        token_path=options.token,
        client_secrets=options.client_secrets,
        port=options.oauth_port,
    )
    return create_canonical_csv_from_gsc_impressions(
        input_csv=options.input_csv,
        output_csv=options.output_csv,
        site_url=options.site_url,
        credentials=credentials,
        interval=options.interval,
        url_regex=options.url_regex,
        concurrency=options.concurrency,
        request_timeout_sec=options.request_timeout_sec,
    )



"""Shared output filename template rendering utilities."""

from __future__ import annotations

import re
from datetime import datetime
from pathlib import Path

_SEQ_PATTERN = re.compile(r"\{seq(?::(\d+))?\}")


def render_output_path_template(
    template: str,
    *,
    profile_name: str,
    now: datetime | None = None,
) -> Path:
    current = now or datetime.now()
    values = {
        "profile": profile_name,
        "yyyy": current.strftime("%Y"),
        "MM": current.strftime("%m"),
        "dd": current.strftime("%d"),
        "HH": current.strftime("%H"),
        "mm": current.strftime("%M"),
        "ss": current.strftime("%S"),
        "date": current.strftime("%Y%m%d"),
        "time": current.strftime("%H%M%S"),
    }
    seq_widths: list[int] = []

    def _seq_replacer(match: re.Match[str]) -> str:
        width = int(match.group(1) or "1")
        token_idx = len(seq_widths)
        seq_widths.append(width)
        return f"__WORAI_SEQ_{token_idx}__"

    protected = _SEQ_PATTERN.sub(_seq_replacer, template)
    base = protected.format(**values)
    if not seq_widths:
        return Path(base)

    seq = 1
    while True:
        rendered = base
        for idx, width in enumerate(seq_widths):
            rendered = rendered.replace(
                f"__WORAI_SEQ_{idx}__",
                str(seq).zfill(width),
            )
        candidate = Path(rendered)
        if not candidate.exists():
            return candidate
        seq += 1



"""Export graph data from WordLift /dataset/export endpoint."""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

import requests
from wordlift_sdk.validation import validate_file

DEFAULT_EXPORT_ENDPOINT = "https://api.wordlift.io/dataset/export"
_DEFAULT_TIMEOUT = 120

_EXTENSION_TO_ACCEPT = {
    ".ttl": "text/turtle",
    ".nt": "application/n-triples",
    ".nq": "application/n-quads",
    ".rdf": "application/rdf+xml",
    ".xml": "application/rdf+xml",
    ".jsonld": "application/ld+json",
    ".json": "application/ld+json",
}


@dataclass
class GraphExportOptions:
    api_key: str
    output_path: Path
    endpoint: str = DEFAULT_EXPORT_ENDPOINT
    timeout: int = _DEFAULT_TIMEOUT


def resolve_accept_header(path: Path) -> str:
    ext = path.suffix.lower()
    accept = _EXTENSION_TO_ACCEPT.get(ext)
    if accept is None:
        supported = ", ".join(sorted(_EXTENSION_TO_ACCEPT))
        raise ValueError(f"Unsupported output extension '{ext or '<none>'}'. Supported: {supported}.")
    return accept


def run_graph_export(options: GraphExportOptions) -> Path:
    accept = resolve_accept_header(options.output_path)
    headers = {
        "Authorization": f"Key {options.api_key}",
        "Accept": accept,
    }
    try:
        response = requests.get(options.endpoint, headers=headers, timeout=options.timeout)
    except requests.RequestException as exc:
        raise RuntimeError(f"Export request failed: {exc}") from exc

    if response.status_code != 200:
        raise RuntimeError(f"Export HTTP {response.status_code}: {response.text[:500]}")

    output_path = options.output_path
    output_path.parent.mkdir(parents=True, exist_ok=True)
    output_path.write_bytes(response.content)
    return output_path


def validate_exported_graph(path: Path) -> None:
    try:
        result = validate_file(str(path), None)
    except Exception as exc:  # pragma: no cover - sdk-specific runtime details
        raise RuntimeError(f"Failed to run SHACL validation: {exc}") from exc

    if not result.conforms:
        raise RuntimeError("Exported graph failed SHACL validation.")
    if result.warning_count:
        raise RuntimeError("Exported graph contains SHACL warnings.")
